from . import base58
from . import hashes
from .misc import secp256k1
from .networks import NETWORKS
from .base import EmbitBase, EmbitError, EmbitKey
from binascii import hexlify, unhexlify


class ECError(EmbitError):
    pass


class Signature(EmbitBase):
    def __init__(self, sig):
        self._sig = sig

    def write_to(self, stream) -> int:
        return stream.write(secp256k1.ecdsa_signature_serialize_der(self._sig))

    @classmethod
    def read_from(cls, stream):
        der = stream.read(2)
        der += stream.read(der[1])
        return cls(secp256k1.ecdsa_signature_parse_der(der))


class SchnorrSig(EmbitBase):
    def __init__(self, sig):
        assert len(sig) == 64
        self._sig = sig

    def write_to(self, stream) -> int:
        return stream.write(self._sig)

    @classmethod
    def read_from(cls, stream):
        return cls(stream.read(64))


class PublicKey(EmbitKey):
    def __init__(self, point: bytes, compressed: bool = True):
        self._point = point
        self.compressed = compressed

    @classmethod
    def read_from(cls, stream):
        b = stream.read(1)
        if b not in [b"\x02", b"\x03", b"\x04"]:
            raise ECError("Invalid public key")
        if b == b"\x04":
            b += stream.read(64)
        else:
            b += stream.read(32)
        try:
            point = secp256k1.ec_pubkey_parse(b)
        except Exception as e:
            raise ECError(str(e))
        compressed = b[0] != 0x04
        return cls(point, compressed)

    def sec(self) -> bytes:
        """Sec representation of the key"""
        flag = secp256k1.EC_COMPRESSED if self.compressed else secp256k1.EC_UNCOMPRESSED
        return secp256k1.ec_pubkey_serialize(self._point, flag)

    def xonly(self) -> bytes:
        return self.sec()[1:33]

    def taproot_tweak(self, h=b""):
        """Returns a tweaked public key"""
        x = self.xonly()
        tweak = hashes.tagged_hash("TapTweak", x + h)
        if not secp256k1.ec_seckey_verify(tweak):
            raise EmbitError("Tweak is too large")
        point = secp256k1.ec_pubkey_parse(b"\x02" + x)
        pub = secp256k1.ec_pubkey_add(point, tweak)
        sec = secp256k1.ec_pubkey_serialize(pub)
        return PublicKey.from_xonly(sec[1:33])

    def write_to(self, stream) -> int:
        return stream.write(self.sec())

    def serialize(self) -> bytes:
        return self.sec()

    def verify(self, sig, msg_hash) -> bool:
        return bool(secp256k1.ecdsa_verify(sig._sig, msg_hash, self._point))

    def _xonly(self):
        """Returns internal representation of the xonly-pubkey (64 bytes)"""
        pub, _ = secp256k1.xonly_pubkey_from_pubkey(self._point)
        return pub

    @classmethod
    def from_xonly(cls, data: bytes):
        assert len(data) == 32
        return cls.parse(b"\x02" + data)

    def schnorr_verify(self, sig, msg_hash) -> bool:
        return bool(secp256k1.schnorrsig_verify(sig._sig, msg_hash, self._xonly()))

    @classmethod
    def from_string(cls, s):
        return cls.parse(unhexlify(s))

    @property
    def is_private(self) -> bool:
        return False

    def to_string(self):
        return hexlify(self.sec()).decode()

    def __lt__(self, other):
        # for lexagraphic ordering
        return self.sec() < other.sec()

    def __gt__(self, other):
        # for lexagraphic ordering
        return self.sec() > other.sec()

    def __eq__(self, other):
        return self.sec() == other.sec()

    def __hash__(self):
        return hash(self._point)


class PrivateKey(EmbitKey):
    def __init__(self, secret, compressed: bool = True, network=NETWORKS["main"]):
        """Creates a private key from 32-byte array"""
        if len(secret) != 32:
            raise ECError("Secret should be 32-byte array")
        if not secp256k1.ec_seckey_verify(secret):
            raise ECError("Secret is not valid (larger then N?)")
        self.compressed = compressed
        self._secret = secret
        self.network = network

    def wif(self, network=None) -> str:
        """Export private key as Wallet Import Format string.
        Prefix 0x80 is used for mainnet, 0xEF for testnet.
        This class doesn't store this information though.
        """
        if network is None:
            network = self.network
        prefix = network["wif"]
        b = prefix + self._secret
        if self.compressed:
            b += bytes([0x01])
        return base58.encode_check(b)

    @property
    def secret(self):
        return self._secret

    def sec(self) -> bytes:
        """Sec representation of the corresponding public key"""
        return self.get_public_key().sec()

    def xonly(self) -> bytes:
        return self.sec()[1:]

    def taproot_tweak(self, h=b""):
        """Returns a tweaked private key"""
        sec = self.sec()
        negate = sec[0] != 0x02
        x = sec[1:33]
        tweak = hashes.tagged_hash("TapTweak", x + h)
        if not secp256k1.ec_seckey_verify(tweak):
            raise EmbitError("Tweak is too large")
        if negate:
            secret = secp256k1.ec_privkey_negate(self._secret)
        else:
            secret = self._secret
        res = secp256k1.ec_privkey_add(secret, tweak)
        pk = PrivateKey(res)
        if pk.sec()[0] == 0x03:
            pk = PrivateKey(secp256k1.ec_privkey_negate(res))
        return pk

    @classmethod
    def from_wif(cls, s):
        """Import private key from Wallet Import Format string."""
        b = base58.decode_check(s)
        prefix = b[:1]
        network = None
        for net in NETWORKS:
            if NETWORKS[net]["wif"] == prefix:
                network = NETWORKS[net]
        secret = b[1:33]
        compressed = False
        if len(b) not in [33, 34]:
            raise ECError("Wrong WIF length")
        if len(b) == 34:
            if b[-1] == 0x01:
                compressed = True
            else:
                raise ECError("Wrong WIF compressed flag")
        return cls(secret, compressed, network)

    # to unify API
    def to_base58(self, network=None) -> str:
        return self.wif(network)

    @classmethod
    def from_base58(cls, s):
        return cls.from_wif(s)

    def get_public_key(self) -> PublicKey:
        return PublicKey(secp256k1.ec_pubkey_create(self._secret), self.compressed)

    def to_public(self) -> PublicKey:
        """Alias to get_public_key for API consistency"""
        return self.get_public_key()

    def sign(self, msg_hash, grind=True) -> Signature:
        sig = Signature(secp256k1.ecdsa_sign(msg_hash, self._secret))
        if grind:
            counter = 1
            while len(sig.serialize()) > 70:
                sig = Signature(
                    secp256k1.ecdsa_sign(
                        msg_hash, self._secret, None, counter.to_bytes(32, "little")
                    )
                )
                counter += 1
                # just in case we get in infinite loop for some reason
                if counter > 200:
                    break
        return sig

    def schnorr_sign(self, msg_hash) -> SchnorrSig:
        return SchnorrSig(secp256k1.schnorrsig_sign(msg_hash, self._secret))

    def verify(self, sig, msg_hash) -> bool:
        return self.get_public_key().verify(sig, msg_hash)

    def schnorr_verify(self, sig, msg_hash) -> bool:
        return self.get_public_key().schnorr_verify(sig, msg_hash)

    def write_to(self, stream) -> int:
        # return a copy of the secret
        return stream.write(self._secret)

    def ecdh(self, public_key: PublicKey, hashfn=None, data=None) -> bytes:
        pubkey_point = secp256k1.ec_pubkey_parse(public_key.sec())
        return secp256k1.ecdh(pubkey_point, self._secret, hashfn, data)

    @classmethod
    def read_from(cls, stream):
        # just to unify the API
        return cls(stream.read(32))

    @property
    def is_private(self) -> bool:
        return True


# Nothing up my sleeve point for no-internal-key taproot
# see https://github.com/bitcoin/bips/blob/master/bip-0341.mediawiki#constructing-and-spending-taproot-outputs
NUMS_PUBKEY = PublicKey.from_string(
    "0250929b74c1a04954b78b4b6035e97a5e078a5a0f28ec96d547bfee9ace803ac0"
)



# Original version: https://github.com/bitcoin-core/HWI/blob/3fe369d0379212fae1c72729a179d133b0adc872/hwilib/key.py
# Distributed under the MIT License.

"""
PSBT Classes and Utilities
**************************
"""

import base64
import struct

from io import BytesIO, BufferedReader
from typing import (
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from .key import KeyOriginInfo
from .errors import PSBTSerializationError
from .tx import (
    COutPoint,
    CTransaction,
    CTxIn,
    CTxInWitness,
    CTxOut,
)
from ._serialize import (
    deser_compact_size,
    deser_string,
    Readable,
    ser_compact_size,
    ser_string,
    ser_uint256,
    uint256_from_str,
)

def DeserializeHDKeypath(
    f: Readable,
    key: bytes,
    hd_keypaths: MutableMapping[bytes, KeyOriginInfo],
    expected_sizes: Sequence[int],
) -> None:
    """
    :meta private:

    Deserialize a serialized PSBT public key and keypath key-value pair.

    :param f: The byte stream to read the value from.
    :param key: The bytes of the key of the key-value pair.
    :param hd_keypaths: Dictionary of public key bytes to their :class:`~hwilib.key.KeyOriginInfo`.
    :param expected_sizes: List of key lengths expected for the keypair being deserialized.
    """
    if len(key) not in expected_sizes:
        raise PSBTSerializationError("Size of key was not the expected size for the type partial signature pubkey. Length: {}".format(len(key)))
    pubkey = key[1:]
    if pubkey in hd_keypaths:
        raise PSBTSerializationError("Duplicate key, input partial signature for pubkey already provided")

    hd_keypaths[pubkey] = KeyOriginInfo.deserialize(deser_string(f))

def SerializeHDKeypath(hd_keypaths: Mapping[bytes, KeyOriginInfo], type: bytes) -> bytes:
    """
    :meta private:

    Serialize a public key to :class:`~hwilib.key.KeyOriginInfo` mapping as a PSBT key-value pair.

    :param hd_keypaths: The mapping of public key to keypath
    :param type: The PSBT type bytes to use
    :returns: The serialized keypaths
    """
    r = b""
    for pubkey, path in sorted(hd_keypaths.items()):
        r += ser_string(type + pubkey)
        packed = path.serialize()
        r += ser_string(packed)
    return r

class PartiallySignedInput:
    """
    An object for a PSBT input map.
    """

    PSBT_IN_NON_WITNESS_UTXO = 0x00
    PSBT_IN_WITNESS_UTXO = 0x01
    PSBT_IN_PARTIAL_SIG = 0x02
    PSBT_IN_SIGHASH_TYPE = 0x03
    PSBT_IN_REDEEM_SCRIPT = 0x04
    PSBT_IN_WITNESS_SCRIPT = 0x05
    PSBT_IN_BIP32_DERIVATION = 0x06
    PSBT_IN_FINAL_SCRIPTSIG = 0x07
    PSBT_IN_FINAL_SCRIPTWITNESS = 0x08
    PSBT_IN_PREVIOUS_TXID = 0x0e
    PSBT_IN_OUTPUT_INDEX = 0x0f
    PSBT_IN_SEQUENCE = 0x10
    PSBT_IN_REQUIRED_TIME_LOCKTIME = 0x11
    PSBT_IN_REQUIRED_HEIGHT_LOCKTIME = 0x12
    PSBT_IN_TAP_KEY_SIG = 0x13
    PSBT_IN_TAP_SCRIPT_SIG = 0x14
    PSBT_IN_TAP_LEAF_SCRIPT = 0x15
    PSBT_IN_TAP_BIP32_DERIVATION = 0x16
    PSBT_IN_TAP_INTERNAL_KEY = 0x17
    PSBT_IN_TAP_MERKLE_ROOT = 0x18

    def __init__(self, version: int) -> None:
        self.non_witness_utxo: Optional[CTransaction] = None
        self.witness_utxo: Optional[CTxOut] = None
        self.partial_sigs: Dict[bytes, bytes] = {}
        self.sighash: Optional[int] = None
        self.redeem_script = b""
        self.witness_script = b""
        self.hd_keypaths: Dict[bytes, KeyOriginInfo] = {}
        self.final_script_sig = b""
        self.final_script_witness = CTxInWitness()
        self.prev_txid = b""
        self.prev_out: Optional[int] = None
        self.sequence: Optional[int] = None
        self.time_locktime: Optional[int] = None
        self.height_locktime: Optional[int] = None
        self.tap_key_sig = b""
        self.tap_script_sigs: Dict[Tuple[bytes, bytes], bytes] = {}
        self.tap_scripts: Dict[Tuple[bytes, int], Set[bytes]] = {}
        self.tap_bip32_paths: Dict[bytes, Tuple[Set[bytes], KeyOriginInfo]] = {}
        self.tap_internal_key = b""
        self.tap_merkle_root = b""
        self.unknown: Dict[bytes, bytes] = {}

        self.version: int = version

    def set_null(self) -> None:
        """
        Clear all values in this PSBT input map.
        """
        self.non_witness_utxo = None
        self.witness_utxo = None
        self.partial_sigs.clear()
        self.sighash = None
        self.redeem_script = b""
        self.witness_script = b""
        self.hd_keypaths.clear()
        self.final_script_sig = b""
        self.final_script_witness = CTxInWitness()
        self.tap_key_sig = b""
        self.tap_script_sigs.clear()
        self.tap_scripts.clear()
        self.tap_bip32_paths.clear()
        self.tap_internal_key = b""
        self.tap_merkle_root = b""
        self.prev_txid = b""
        self.prev_out = None
        self.sequence = None
        self.time_locktime = None
        self.height_locktime = None
        self.unknown.clear()

    def deserialize(self, f: Readable) -> None:
        """
        Deserialize a serialized PSBT input.

        :param f: A byte stream containing the serialized PSBT input
        """
        key_lookup: Set[bytes] = set()

        while True:
            # read the key
            try:
                key = deser_string(f)
            except Exception:
                break

            # Check for separator
            if len(key) == 0:
                break

            # First byte of key is the type
            key_type = deser_compact_size(BytesIO(key))

            if key_type == PartiallySignedInput.PSBT_IN_NON_WITNESS_UTXO:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate Key, input non witness utxo already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("non witness utxo key is more than one byte type")
                self.non_witness_utxo = CTransaction()
                utxo_bytes = BufferedReader(BytesIO(deser_string(f))) # type: ignore
                self.non_witness_utxo.deserialize(utxo_bytes)
                self.non_witness_utxo.rehash()
            elif key_type == PartiallySignedInput.PSBT_IN_WITNESS_UTXO:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate Key, input witness utxo already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("witness utxo key is more than one byte type")
                self.witness_utxo = CTxOut()
                tx_out_bytes = BufferedReader(BytesIO(deser_string(f))) # type: ignore
                self.witness_utxo.deserialize(tx_out_bytes)
            elif key_type == PartiallySignedInput.PSBT_IN_PARTIAL_SIG:
                if len(key) != 34 and len(key) != 66:
                    raise PSBTSerializationError("Size of key was not the expected size for the type partial signature pubkey")
                pubkey = key[1:]
                if pubkey in self.partial_sigs:
                    raise PSBTSerializationError("Duplicate key, input partial signature for pubkey already provided")

                sig = deser_string(f)
                self.partial_sigs[pubkey] = sig
            elif key_type == PartiallySignedInput.PSBT_IN_SIGHASH_TYPE:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input sighash type already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("sighash key is more than one byte type")
                sighash_bytes = deser_string(f)
                self.sighash = struct.unpack("<I", sighash_bytes)[0]
            elif key_type == PartiallySignedInput.PSBT_IN_REDEEM_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input redeemScript already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("redeemScript key is more than one byte type")
                self.redeem_script = deser_string(f)
            elif key_type == PartiallySignedInput.PSBT_IN_WITNESS_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input witnessScript already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("witnessScript key is more than one byte type")
                self.witness_script = deser_string(f)
            elif key_type == PartiallySignedInput.PSBT_IN_BIP32_DERIVATION:
                DeserializeHDKeypath(f, key, self.hd_keypaths, [34, 66])
            elif key_type == PartiallySignedInput.PSBT_IN_FINAL_SCRIPTSIG:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input final scriptSig already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("final scriptSig key is more than one byte type")
                self.final_script_sig = deser_string(f)
            elif key_type == PartiallySignedInput.PSBT_IN_FINAL_SCRIPTWITNESS:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input final scriptWitness already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("final scriptWitness key is more than one byte type")
                witness_bytes = BufferedReader(BytesIO(deser_string(f))) # type: ignore
                self.final_script_witness.deserialize(witness_bytes)
            elif key_type == PartiallySignedInput.PSBT_IN_PREVIOUS_TXID:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input previous txid is already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Previous txid key is more than one byte type")
                txid = deser_string(f)
                if len(txid) != 32:
                    raise PSBTSerializationError("Previous txid is not 32 bytes")
                self.prev_txid = txid
            elif key_type == PartiallySignedInput.PSBT_IN_OUTPUT_INDEX:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input previous output index is already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Previous output index key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Previous output index is not 4 bytes")
                self.prev_out = struct.unpack("<I", v)[0]
            elif key_type == PartiallySignedInput.PSBT_IN_SEQUENCE:
                pass
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input sequence is already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input sequence key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Input sequence is not 4 bytes")
                self.sequence = struct.unpack("<I", v)[0]
            elif key_type == PartiallySignedInput.PSBT_IN_REQUIRED_TIME_LOCKTIME:
                pass
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input required time based locktime is already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input time based locktime key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Input time based locktime is not 4 bytes")
                self.time_locktime = struct.unpack("<I", v)[0]
            elif key_type == PartiallySignedInput.PSBT_IN_REQUIRED_HEIGHT_LOCKTIME:
                pass
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input required height based locktime index is already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input height based locktime key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Input height based locktime is not 4 bytes")
                self.height_locktime = struct.unpack("<I", v)[0]
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_KEY_SIG:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot key signature already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input Taproot key signature key is more than one byte type")
                self.tap_key_sig = deser_string(f)
                if len(self.tap_key_sig) < 64:
                    raise PSBTSerializationError("Input Taproot key path signature is shorter than 64 bytes")
                elif len(self.tap_key_sig) > 65:
                    raise PSBTSerializationError("Input Taproot key path signature is longer than 65 bytes")
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_SCRIPT_SIG:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot script signature already provided")
                elif len(key) != 65:
                    raise PSBTSerializationError("Input Taproot script signature key is not 65 bytes")
                xonly = key[1:33]
                script_hash = key[33:65]
                sig = deser_string(f)
                if len(sig) < 64:
                    raise PSBTSerializationError("Input Taproot script path signature is shorter than 64 bytes")
                elif len(sig) > 65:
                    raise PSBTSerializationError("Input Taproot script path signature is longer than 65 bytes")
                self.tap_script_sigs[(xonly, script_hash)] = sig
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_LEAF_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot leaf script already provided")
                elif len(key) < 34:
                    raise PSBTSerializationError("Input Taproot leaf script key is not at least 34 bytes")
                elif (len(key) - 2) % 32 != 0:
                    raise PSBTSerializationError("Input Taproot leaf script key's control block is not valid")
                script = deser_string(f)
                if len(script) == 0:
                    raise PSBTSerializationError("Input Taproot leaf script cannot be empty")
                leaf_script = (script[:-1], int(script[-1]))
                if leaf_script not in self.tap_scripts:
                    self.tap_scripts[leaf_script] = set()
                self.tap_scripts[(script[:-1], int(script[-1]))].add(key[1:])
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_BIP32_DERIVATION:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot BIP 32 keypath already provided")
                elif len(key) != 33:
                    raise PSBTSerializationError("Input Taproot BIP 32 keypath key is not 33 bytes")
                xonly = key[1:33]
                value = deser_string(f)
                vs = BytesIO(value)
                num_hashes = deser_compact_size(vs)
                leaf_hashes = set()
                for i in range(0, num_hashes):
                    leaf_hashes.add(vs.read(32))
                self.tap_bip32_paths[xonly] = (leaf_hashes, KeyOriginInfo.deserialize(vs.read()))
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_INTERNAL_KEY:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot internal key already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input Taproot internal key key is more than one byte type")
                self.tap_internal_key = deser_string(f)
                if len(self.tap_internal_key) != 32:
                    raise PSBTSerializationError("Input Taproot internal key is not 32 bytes")
            elif key_type == PartiallySignedInput.PSBT_IN_TAP_MERKLE_ROOT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, input Taproot merkle root already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Input Taproot merkle root key is more than one byte type")
                self.tap_merkle_root = deser_string(f)
                if len(self.tap_merkle_root) != 32:
                    raise PSBTSerializationError("Input Taproot merkle root is not 32 bytes")
            else:
                if key in self.unknown:
                    raise PSBTSerializationError("Duplicate key, key for unknown value already provided")
                unknown_bytes = deser_string(f)
                self.unknown[key] = unknown_bytes

            key_lookup.add(key)

        # Make sure required PSBTv2 fields are present
        if self.version >= 2:
            if len(self.prev_txid) == 0:
                raise PSBTSerializationError("Previous TXID is required in PSBTv2")
            if self.prev_out is None:
                raise PSBTSerializationError("Previous output's index is required in PSBTv2")

    def serialize(self) -> bytes:
        """
        Serialize this PSBT input

        :returns: The serialized PSBT input
        """
        r = b""

        if self.non_witness_utxo:
            r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_NON_WITNESS_UTXO))
            tx = self.non_witness_utxo.serialize_with_witness()
            r += ser_string(tx)

        if self.witness_utxo:
            r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_WITNESS_UTXO))
            tx = self.witness_utxo.serialize()
            r += ser_string(tx)

        if len(self.final_script_sig) == 0 and self.final_script_witness.is_null():
            for pubkey, sig in sorted(self.partial_sigs.items()):
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_PARTIAL_SIG) + pubkey)
                r += ser_string(sig)

            if self.sighash is not None:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_SIGHASH_TYPE))
                r += ser_string(struct.pack("<I", self.sighash))

            if len(self.redeem_script) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_REDEEM_SCRIPT))
                r += ser_string(self.redeem_script)

            if len(self.witness_script) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_WITNESS_SCRIPT))
                r += ser_string(self.witness_script)

            r += SerializeHDKeypath(self.hd_keypaths, ser_compact_size(PartiallySignedInput.PSBT_IN_BIP32_DERIVATION))

            if len(self.tap_key_sig) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_KEY_SIG))
                r += ser_string(self.tap_key_sig)

            for (xonly, leaf_hash), sig in self.tap_script_sigs.items():
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_SCRIPT_SIG) + xonly + leaf_hash)
                r += ser_string(sig)

            for (script, leaf_ver), control_blocks in self.tap_scripts.items():
                for control_block in control_blocks:
                    r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_LEAF_SCRIPT) + control_block)
                    r += ser_string(script + struct.pack("B", leaf_ver))

            for xonly, (leaf_hashes, origin) in self.tap_bip32_paths.items():
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_BIP32_DERIVATION) + xonly)
                value = ser_compact_size(len(leaf_hashes))
                for lh in leaf_hashes:
                    value += lh
                value += origin.serialize()
                r += ser_string(value)

            if len(self.tap_internal_key) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_INTERNAL_KEY))
                r += ser_string(self.tap_internal_key)

            if len(self.tap_merkle_root) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_TAP_MERKLE_ROOT))
                r += ser_string(self.tap_merkle_root)

        if len(self.final_script_sig) != 0:
            r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_FINAL_SCRIPTSIG))
            r += ser_string(self.final_script_sig)

        if not self.final_script_witness.is_null():
            r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_FINAL_SCRIPTWITNESS))
            witstack = self.final_script_witness.serialize()
            r += ser_string(witstack)

        if self.version >= 2:
            if len(self.prev_txid) != 0:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_PREVIOUS_TXID))
                r += ser_string(self.prev_txid)

            if self.prev_out is not None:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_OUTPUT_INDEX))
                r += ser_string(struct.pack("<I", self.prev_out))

            if self.sequence is not None:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_SEQUENCE))
                r += ser_string(struct.pack("<I", self.sequence))

            if self.time_locktime is not None:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_REQUIRED_TIME_LOCKTIME))
                r += ser_string(struct.pack("<I", self.time_locktime))

            if self.height_locktime is not None:
                r += ser_string(ser_compact_size(PartiallySignedInput.PSBT_IN_REQUIRED_HEIGHT_LOCKTIME))
                r += ser_string(struct.pack("<I", self.height_locktime))

        for key, value in sorted(self.unknown.items()):
            r += ser_string(key)
            r += ser_string(value)

        r += b"\x00"

        return r

class PartiallySignedOutput:
    """
    An object for a PSBT output map.
    """

    PSBT_OUT_REDEEM_SCRIPT = 0x00
    PSBT_OUT_WITNESS_SCRIPT = 0x01
    PSBT_OUT_BIP32_DERIVATION = 0x02
    PSBT_OUT_AMOUNT = 0x03
    PSBT_OUT_SCRIPT = 0x04
    PSBT_OUT_TAP_INTERNAL_KEY = 0x05
    PSBT_OUT_TAP_TREE = 0x06
    PSBT_OUT_TAP_BIP32_DERIVATION = 0x07

    def __init__(self, version: int) -> None:
        self.redeem_script = b""
        self.witness_script = b""
        self.hd_keypaths: Dict[bytes, KeyOriginInfo] = {}
        self.amount: Optional[int] = None
        self.script = b""
        self.tap_internal_key = b""
        self.tap_tree = b""
        self.tap_bip32_paths: Dict[bytes, Tuple[Set[bytes], KeyOriginInfo]] = {}
        self.unknown: Dict[bytes, bytes] = {}

        self.version: int = version

    def set_null(self) -> None:
        """
        Clear this PSBT output map
        """
        self.redeem_script = b""
        self.witness_script = b""
        self.hd_keypaths.clear()
        self.tap_internal_key = b""
        self.tap_tree = b""
        self.tap_bip32_paths.clear()
        self.amount = None
        self.script = b""
        self.unknown.clear()

    def deserialize(self, f: Readable) -> None:
        """
        Deserialize a serialized PSBT output map

        :param f: A byte stream containing the serialized PSBT output
        """
        key_lookup: Set[bytes] = set()

        while True:
            # read the key
            try:
                key = deser_string(f)
            except Exception:
                break

            # Check for separator
            if len(key) == 0:
                break

            # First byte of key is the type
            key_type = deser_compact_size(BytesIO(key))

            if key_type == PartiallySignedOutput.PSBT_OUT_REDEEM_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output redeemScript already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output redeemScript key is more than one byte type")
                self.redeem_script = deser_string(f)
            elif key_type == PartiallySignedOutput.PSBT_OUT_WITNESS_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output witnessScript already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output witnessScript key is more than one byte type")
                self.witness_script = deser_string(f)
            elif key_type == PartiallySignedOutput.PSBT_OUT_BIP32_DERIVATION:
                DeserializeHDKeypath(f, key, self.hd_keypaths, [34, 66])
            elif key_type == PartiallySignedOutput.PSBT_OUT_AMOUNT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output amount already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output amount key is more than one byte type")
                v = deser_string(f)
                if len(v) != 8:
                    raise PSBTSerializationError("Output amount is not 8 bytes")
                self.amount = struct.unpack("<q", v)[0]
            elif key_type == PartiallySignedOutput.PSBT_OUT_SCRIPT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output script already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output script key is more than one byte type")
                self.script = deser_string(f)
            elif key_type == PartiallySignedOutput.PSBT_OUT_TAP_INTERNAL_KEY:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output Taproot internal key already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output Taproot internal key key is more than one byte type")
                self.tap_internal_key = deser_string(f)
                if len(self.tap_internal_key) != 32:
                    raise PSBTSerializationError("Output Taproot internal key is not 32 bytes")
            elif key_type == PartiallySignedOutput.PSBT_OUT_TAP_TREE:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output Taproot tree already provided")
                elif len(key) != 1:
                    raise PSBTSerializationError("Output Taproot tree key is more than one byte type")
                self.tap_tree = deser_string(f)
            elif key_type == PartiallySignedOutput.PSBT_OUT_TAP_BIP32_DERIVATION:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, output Taproot BIP 32 keypath already provided")
                elif len(key) != 33:
                    raise PSBTSerializationError("Output Taproot BIP 32 keypath key is not 33 bytes")
                xonly = key[1:33]
                value = deser_string(f)
                vs = BytesIO(value)
                num_hashes = deser_compact_size(vs)
                leaf_hashes = set()
                for i in range(0, num_hashes):
                    leaf_hashes.add(vs.read(32))
                self.tap_bip32_paths[xonly] = (leaf_hashes, KeyOriginInfo.deserialize(vs.read()))
            else:
                if key in self.unknown:
                    raise PSBTSerializationError("Duplicate key, key for unknown value already provided")
                value = deser_string(f)
                self.unknown[key] = value

            key_lookup.add(key)

        # Make sure required PSBTv2 fields are present
        if self.version >= 2:
            if self.amount is None:
                raise PSBTSerializationError("PSBT_OUTPUT_AMOUNT is required in PSBTv2")
            if len(self.script) == 0:
                raise PSBTSerializationError("PSBT_OUTPUT_SCRIPT is required in PSBTv2")

    def serialize(self) -> bytes:
        """
        Serialize this PSBT output

        :returns: The serialized PSBT output
        """
        r = b""
        if len(self.redeem_script) != 0:
            r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_REDEEM_SCRIPT))
            r += ser_string(self.redeem_script)

        if len(self.witness_script) != 0:
            r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_WITNESS_SCRIPT))
            r += ser_string(self.witness_script)

        r += SerializeHDKeypath(self.hd_keypaths, ser_compact_size(PartiallySignedOutput.PSBT_OUT_BIP32_DERIVATION))

        if self.version >= 2:
            if self.amount is not None:
                r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_AMOUNT))
                r += ser_string(struct.pack("<q", self.amount))

            if len(self.script) != 0:
                r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_SCRIPT))
                r += ser_string(self.script)

        if len(self.tap_internal_key) != 0:
            r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_TAP_INTERNAL_KEY))
            r += ser_string(self.tap_internal_key)

        if len(self.tap_tree) != 0:
            r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_TAP_TREE))
            r += ser_string(self.tap_tree)

        for xonly, (leaf_hashes, origin) in self.tap_bip32_paths.items():
            r += ser_string(ser_compact_size(PartiallySignedOutput.PSBT_OUT_TAP_BIP32_DERIVATION) + xonly)
            value = ser_compact_size(len(leaf_hashes))
            for lh in leaf_hashes:
                value += lh
            value += origin.serialize()
            r += ser_string(value)

        for key, value in sorted(self.unknown.items()):
            r += ser_string(key)
            r += ser_string(value)

        r += b"\x00"

        return r

    def get_txout(self) -> CTxOut:
        """
        Creates a CTxOut for this output

        :returns: The CTxOut
        """
        assert self.amount is not None
        assert len(self.script) != 0
        return CTxOut(self.amount, self.script)

class PSBT(object):
    """
    A class representing a PSBT
    """

    PSBT_GLOBAL_UNSIGNED_TX = 0x00
    PSBT_GLOBAL_XPUB = 0x01
    PSBT_GLOBAL_TX_VERSION = 0x02
    PSBT_GLOBAL_FALLBACK_LOCKTIME = 0x03
    PSBT_GLOBAL_INPUT_COUNT = 0x04
    PSBT_GLOBAL_OUTPUT_COUNT = 0x05
    PSBT_GLOBAL_TX_MODIFIABLE = 0x06
    PSBT_GLOBAL_VERSION = 0xFB

    def __init__(self, tx: Optional[CTransaction] = None) -> None:
        """
        :param tx: A Bitcoin transaction that specifies the inputs and outputs to use
        """
        if tx:
            self.tx = tx
        else:
            self.tx = CTransaction()
        self.inputs: List[PartiallySignedInput] = []
        self.outputs: List[PartiallySignedOutput] = []
        self.unknown: Dict[bytes, bytes] = {}
        self.xpub: Dict[bytes, KeyOriginInfo] = {}
        self.tx_version: Optional[int] = None
        self.fallback_locktime: Optional[int] = None
        self.tx_modifiable: Optional[int] = None

        # Assume version 0 PSBT
        self.version = 0
        self.explicit_version = False

    def deserialize(self, psbt: str) -> None:
        """
        Deserialize a base 64 encoded PSBT.

        :param psbt: A base 64 PSBT.
        """
        psbt_bytes = base64.b64decode(psbt.strip())
        f = BufferedReader(BytesIO(psbt_bytes)) # type: ignore
        end = len(psbt_bytes)

        # Read the magic bytes
        magic = f.read(5)
        if magic != b"psbt\xff":
            raise PSBTSerializationError("invalid magic")

        key_lookup: Set[bytes] = set()

        input_count = None
        output_count = None

        # Read loop
        while True:
            # read the key
            try:
                key = deser_string(f)
            except Exception:
                break

            # Check for separator
            if len(key) == 0:
                break

            # First byte of key is the type
            key_type = deser_compact_size(BytesIO(key))

            # Do stuff based on type
            if key_type == PSBT.PSBT_GLOBAL_UNSIGNED_TX:
                # Checks for correctness
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, unsigned tx already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global unsigned tx key is more than one byte type")

                # read in value
                tx_bytes = BufferedReader(BytesIO(deser_string(f))) # type: ignore
                self.tx.deserialize(tx_bytes)

                # Make sure that all scriptSigs and scriptWitnesses are empty
                for txin in self.tx.vin:
                    if len(txin.scriptSig) != 0 or not self.tx.wit.is_null():
                        raise PSBTSerializationError("Unsigned tx does not have empty scriptSigs and scriptWitnesses")
            elif key_type == PSBT.PSBT_GLOBAL_XPUB:
                DeserializeHDKeypath(f, key, self.xpub, [79])
            elif key_type == PSBT.PSBT_GLOBAL_TX_VERSION:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global transaction version is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global transaction version key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Global transaction version is not 4 bytes")
                self.tx_version = struct.unpack("<I", v)[0]
            elif key_type == PSBT.PSBT_GLOBAL_FALLBACK_LOCKTIME:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global fallback locktime is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global fallback locktime key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Global fallback locktime is not 4 bytes")
                self.fallback_locktime = struct.unpack("<I", v)[0]
            elif key_type == PSBT.PSBT_GLOBAL_INPUT_COUNT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global input count is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global input count key is more than one byte type")
                _ = deser_compact_size(f) # Value length, we can ignore this
                input_count = deser_compact_size(f)
            elif key_type == PSBT.PSBT_GLOBAL_OUTPUT_COUNT:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global output count is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global output count key is more than one byte type")
                _ = deser_compact_size(f) # Value length, we can ignore this
                output_count = deser_compact_size(f)
            elif key_type == PSBT.PSBT_GLOBAL_TX_MODIFIABLE:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global tx modifiable flags is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global tx modifiable flags key is more than one byte type")
                v = deser_string(f)
                if len(v) != 1:
                    raise PSBTSerializationError("Global tx modifiable flags is not 1 bytes")
                self.tx_modifiable = struct.unpack("<B", v)[0]
            elif key_type == PSBT.PSBT_GLOBAL_VERSION:
                if key in key_lookup:
                    raise PSBTSerializationError("Duplicate key, global PSBT version is already provided")
                elif len(key) > 1:
                    raise PSBTSerializationError("Global PSBT version key is more than one byte type")
                v = deser_string(f)
                if len(v) != 4:
                    raise PSBTSerializationError("Global PSBT version is not 1 bytes")
                self.version = struct.unpack("<I", v)[0]
                self.explicit_version = True
            else:
                if key in self.unknown:
                    raise PSBTSerializationError("Duplicate key, key for unknown value already provided")
                unknown_bytes = deser_string(f)
                self.unknown[key] = unknown_bytes

            key_lookup.add(key)

        # Check PSBT version constraints
        if self.version == 0:
            # make sure that we got an unsigned tx
            if self.tx.is_null():
                raise PSBTSerializationError("No unsigned transaction was provided")
            # Make sure no v2 fields are present
            if self.tx_version is not None:
                raise PSBTSerializationError("PSBT_GLOBAL_TX_VERSION is not allowed in PSBTv0")
            if self.fallback_locktime is not None:
                raise PSBTSerializationError("PSBT_GLOBAL_FALLBACK_LOCKTIME is not allowed in PSBTv0")
            if input_count is not None:
                raise PSBTSerializationError("PSBT_GLOBAL_INPUT_COUNT is not allowed in PSBTv0")
            if output_count is not None:
                raise PSBTSerializationError("PSBT_GLOBAL_OUTPUT_COUNT is not allowed in PSBTv0")
            if self.tx_modifiable is not None:
                raise PSBTSerializationError("PSBT_GLOBAL_TX_MODIFIABLE is not allowed in PSBTv0")

        # Disallow v1
        if self.version == 1:
            raise PSBTSerializationError("There is no PSBT version 1")
        if self.version >= 2:
            # Tx version, input, and output counts are required
            if self.tx_version is None:
                raise PSBTSerializationError("PSBT_GLOBAL_TX_VERSION is required in PSBTv2")
            if input_count is None:
                raise PSBTSerializationError("PSBT_GLOBAL_INPUT_COUNT is required in PSBTv2")
            if output_count is None:
                raise PSBTSerializationError("PSBT_GLOBAL_OUTPUT_COUNT is required in PSBTv2")
            # Unsigned tx is disallowed
            if not self.tx.is_null():
                raise PSBTSerializationError("PSBT_GLOBAL_UNSIGNED_TX is not allowed in PSBTv2")

        # Read input data
        if input_count is None:
            input_count = len(self.tx.vin)
        for i in range(input_count):
            if f.tell() == end:
                break
            psbt_in = PartiallySignedInput(self.version)
            psbt_in.deserialize(f)
            self.inputs.append(psbt_in)

            if self.version >= 2:
                prev_txid = psbt_in.prev_txid
            else:
                prev_txid = ser_uint256(self.tx.vin[i].prevout.hash)

            if psbt_in.non_witness_utxo:
                psbt_in.non_witness_utxo.rehash()
                if psbt_in.non_witness_utxo.hash != prev_txid:
                    raise PSBTSerializationError("Non-witness UTXO does not match outpoint hash")

        if (len(self.inputs) != input_count):
            raise PSBTSerializationError("Inputs provided does not match the number of inputs in transaction")

        # Read output data
        if output_count is None:
            output_count = len(self.tx.vout)
        for i in range(output_count):
            if f.tell() == end:
                break
            output = PartiallySignedOutput(self.version)
            output.deserialize(f)
            self.outputs.append(output)

        if len(self.outputs) != output_count:
            raise PSBTSerializationError("Outputs provided does not match the number of outputs in transaction")

        self.cache_unsigned_tx_pieces()

    def serialize(self) -> str:
        """
        Serialize the PSBT as a base 64 encoded string.

        :returns: The base 64 encoded string.
        """
        r = b""

        # magic bytes
        r += b"psbt\xff"

        if self.version == 0:
            # unsigned tx flag
            r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_UNSIGNED_TX))

            # write serialized tx
            tx = self.tx.serialize_with_witness()
            r += ser_compact_size(len(tx))
            r += tx

        # write xpubs
        r += SerializeHDKeypath(self.xpub, ser_compact_size(PSBT.PSBT_GLOBAL_XPUB))

        if self.version >= 2:
            assert self.tx_version is not None
            r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_TX_VERSION))
            r += ser_string(struct.pack("<I", self.tx_version))

            if self.fallback_locktime is not None:
                r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_FALLBACK_LOCKTIME))
                r += ser_string(struct.pack("<I", self.fallback_locktime))

            r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_INPUT_COUNT))
            r += ser_string(ser_compact_size(len(self.inputs)))

            r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_OUTPUT_COUNT))
            r += ser_string(ser_compact_size(len(self.outputs)))

            if self.tx_modifiable is not None:
                r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_TX_MODIFIABLE))
                r += ser_string(struct.pack("<B", self.tx_modifiable))

        if self.version > 0 or self.explicit_version:
            r += ser_string(ser_compact_size(PSBT.PSBT_GLOBAL_VERSION))
            r += ser_string(struct.pack("<I", self.version))

        # unknowns
        for key, value in sorted(self.unknown.items()):
            r += ser_string(key)
            r += ser_string(value)

        # separator
        r += b"\x00"

        # inputs
        for input in self.inputs:
            r += input.serialize()

        # outputs
        for output in self.outputs:
            r += output.serialize()

        # return hex string
        return base64.b64encode(r).decode()

    def cache_unsigned_tx_pieces(self) -> None:
        """
        If this PSBT is v0, then the global unsigned transaction will be used to fill in the PSBTv2
        fields so that all users of the PSBT classes can use the same PSBTv2 interface regardless
        of PSBT version.

        Does nothing if the PSBT is already v2.
        """
        # To make things easier, we split up the global transaction
        # and use the PSBTv2 fields for PSBTv0
        if self.tx is not None:
            self.setup_from_tx(self.tx)

    def setup_from_tx(self, tx: CTransaction):
        """
        Fills in the PSBTv2 fields for this PSBT given a transaction

        :param tx: The CTransaction to fill from
        """
        self.tx_version = tx.nVersion
        self.fallback_locktime = tx.nLockTime

        for i, txin in enumerate(tx.vin):
            psbt_in = self.inputs[i]

            psbt_in.prev_txid = ser_uint256(txin.prevout.hash)
            psbt_in.prev_out = txin.prevout.n
            psbt_in.sequence = txin.nSequence

        for i, txout in enumerate(tx.vout):
            psbt_out = self.outputs[i]

            psbt_out.amount = txout.nValue
            psbt_out.script = txout.scriptPubKey

    def compute_lock_time(self) -> int:
        """
        Computes the lock time for this transaction

        :returns: The lock time
        """
        time_lock: Optional[int] = 0
        height_lock: Optional[int] = 0

        for psbt_in in self.inputs:
            if psbt_in.time_locktime is not None and psbt_in.height_locktime is None:
                height_lock = None
                if time_lock is None:
                    raise PSBTSerializationError("Cannot require both time and height locktimes")
            elif psbt_in.time_locktime is None and psbt_in.height_locktime is not None:
                time_lock = None
                if height_lock is None:
                    raise PSBTSerializationError("Cannot require both time and height locktimes")

            if psbt_in.time_locktime is not None and time_lock is not None:
                time_lock = max(time_lock, psbt_in.time_locktime)
            if psbt_in.height_locktime is not None and height_lock is not None:
                height_lock = max(height_lock, psbt_in.height_locktime)

        if height_lock is not None and height_lock > 0:
            return height_lock
        if time_lock is not None and time_lock > 0:
            return time_lock
        if self.fallback_locktime is not None:
            return self.fallback_locktime
        return 0

    def get_unsigned_tx(self) -> CTransaction:
        """
        Get the unsigned transaction represented by this PSBT

        :return: A CTransaction
        """
        if not self.tx.is_null():
            return self.tx

        assert self.tx_version is not None

        tx = CTransaction()
        tx.nVersion = self.tx_version
        self.nLockTime = self.compute_lock_time()

        for psbt_in in self.inputs:
            assert psbt_in.prev_txid is not None
            assert psbt_in.prev_out is not None
            assert psbt_in.sequence is not None

            txin = CTxIn(COutPoint(uint256_from_str(psbt_in.prev_txid), psbt_in.prev_out), b"", psbt_in.sequence)
            tx.vin.append(txin)

        for psbt_out in self.outputs:
            assert psbt_out.amount is not None

            txout = CTxOut(psbt_out.amount, psbt_out.script)
            tx.vout.append(txout)

        tx.rehash()
        return tx

    def _convert_version(self, version) -> None:
        self.version = version
        for psbt_in in self.inputs:
            psbt_in.version = version
        for psbt_out in self.outputs:
            psbt_out.version = version

    def convert_to_v2(self) -> None:
        """
        Sets this PSBT to version 2
        """

        if self.version == 0:
            # make sure fields that PSBT version 0 fields are not present
            self.setup_from_tx(self.tx)

        self._convert_version(2)

    def convert_to_v0(self) -> None:
        """
        Sets this PSBT to version 0
        """

        if self.version == 2:
            # strip PSBT version 2 fields
            self.tx_version = None
            self.fallback_locktime = None
            self.tx_modifiable = None

        self._convert_version(0)
        self.tx = self.get_unsigned_tx()
        self.explicit_version = False


def normalize_psbt(psbt: Union[PSBT, bytes, str]) -> PSBT:
    """
    Deserializes a psbt given as an argument from a string or a byte array, if necessary.

    :param psbt: Either an instance of PSBT, or binary-encoded psbt as `bytes`, or a base64-encoded psbt as a `str`.
    :returns: the deserialized PSBT object. If `psbt` was already a `PSBT`, it is returned directly (without cloning).
    """
    if isinstance(psbt, bytes):
        psbt = base64.b64encode(psbt).decode()

    if isinstance(psbt, str):
        psbt_obj = PSBT()
        psbt_obj.deserialize(psbt)
        psbt = psbt_obj
    return psbt



"""
This is a fallback option if the library can't do ctypes bindings to secp256k1 library.
Mimics the micropython bindings and internal representation of data structs in secp256k1.
"""

from . import key as _key

# Flags to pass to context_create.
CONTEXT_VERIFY = 0b0100000001
CONTEXT_SIGN = 0b1000000001
CONTEXT_NONE = 0b0000000001

# Flags to pass to ec_pubkey_serialize
EC_COMPRESSED = 0b0100000010
EC_UNCOMPRESSED = 0b0000000010


def context_randomize(seed, context=None):
    pass


def _reverse64(b):
    """Converts (a,b) from big to little endian to be consistent with secp256k1"""
    x = b[:32]
    y = b[32:]
    return x[::-1] + y[::-1]


def _pubkey_serialize(pub):
    """Returns pubkey representation like secp library"""
    b = pub.get_bytes()[1:]
    return _reverse64(b)


def _pubkey_parse(b):
    """Returns pubkey class instance"""
    pub = _key.ECPubKey()
    pub.set(b"\x04" + _reverse64(b))
    return pub


def ec_pubkey_create(secret, context=None):
    if len(secret) != 32:
        raise ValueError("Private key should be 32 bytes long")
    pk = _key.ECKey()
    pk.set(secret, compressed=False)
    if not pk.is_valid:
        raise ValueError("Invalid private key")
    return _pubkey_serialize(pk.get_pubkey())


def ec_pubkey_parse(sec, context=None):
    if len(sec) != 33 and len(sec) != 65:
        raise ValueError("Serialized pubkey should be 33 or 65 bytes long")
    if len(sec) == 33:
        if sec[0] != 0x02 and sec[0] != 0x03:
            raise ValueError("Compressed pubkey should start with 0x02 or 0x03")
    else:
        if sec[0] != 0x04:
            raise ValueError("Uncompressed pubkey should start with 0x04")
    pub = _key.ECPubKey()
    pub.set(sec)
    pub.compressed = False
    if not pub.is_valid:
        raise ValueError("Failed parsing public key")
    return _pubkey_serialize(pub)


def ec_pubkey_serialize(pubkey, flag=EC_COMPRESSED, context=None):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be 64 bytes long")
    if flag not in [EC_COMPRESSED, EC_UNCOMPRESSED]:
        raise ValueError("Invalid flag")
    pub = _pubkey_parse(pubkey)
    if not pub.is_valid:
        raise ValueError("Failed to serialize pubkey")
    if flag == EC_COMPRESSED:
        pub.compressed = True
    return pub.get_bytes()


def ecdsa_signature_parse_compact(compact_sig, context=None):
    if len(compact_sig) != 64:
        raise ValueError("Compact signature should be 64 bytes long")
    sig = _reverse64(compact_sig)
    return sig


def ecdsa_signature_parse_der(der, context=None):
    if der[1] + 2 != len(der):
        raise ValueError("Failed parsing compact signature")
    if len(der) < 4:
        raise ValueError("Failed parsing compact signature")
    if der[0] != 0x30:
        raise ValueError("Failed parsing compact signature")
    if der[2] != 0x02:
        raise ValueError("Failed parsing compact signature")
    rlen = der[3]
    if len(der) < 6 + rlen:
        raise ValueError("Failed parsing compact signature")
    if rlen < 1 or rlen > 33:
        raise ValueError("Failed parsing compact signature")
    if der[4] >= 0x80:
        raise ValueError("Failed parsing compact signature")
    if rlen > 1 and (der[4] == 0) and not (der[5] & 0x80):
        raise ValueError("Failed parsing compact signature")
    r = int.from_bytes(der[4 : 4 + rlen], "big")
    if der[4 + rlen] != 0x02:
        raise ValueError("Failed parsing compact signature")
    slen = der[5 + rlen]
    if slen < 1 or slen > 33:
        raise ValueError("Failed parsing compact signature")
    if len(der) != 6 + rlen + slen:
        raise ValueError("Failed parsing compact signature")
    if der[6 + rlen] >= 0x80:
        raise ValueError("Failed parsing compact signature")
    if slen > 1 and (der[6 + rlen] == 0) and not (der[7 + rlen] & 0x80):
        raise ValueError("Failed parsing compact signature")
    s = int.from_bytes(der[6 + rlen : 6 + rlen + slen], "big")

    # Verify that r and s are within the group order
    if r < 1 or s < 1 or r >= _key.SECP256K1_ORDER or s >= _key.SECP256K1_ORDER:
        raise ValueError("Failed parsing compact signature")
    if s >= _key.SECP256K1_ORDER_HALF:
        raise ValueError("Failed parsing compact signature")

    return r.to_bytes(32, "little") + s.to_bytes(32, "little")


def ecdsa_signature_serialize_der(sig, context=None):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    r = int.from_bytes(sig[:32], "little")
    s = int.from_bytes(sig[32:], "little")
    rb = r.to_bytes((r.bit_length() + 8) // 8, "big")
    sb = s.to_bytes((s.bit_length() + 8) // 8, "big")
    return (
        b"\x30"
        + bytes([4 + len(rb) + len(sb), 2, len(rb)])
        + rb
        + bytes([2, len(sb)])
        + sb
    )


def ecdsa_signature_serialize_compact(sig, context=None):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    return _reverse64(sig)


def ecdsa_signature_normalize(sig, context=None):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    r = int.from_bytes(sig[:32], "little")
    s = int.from_bytes(sig[32:], "little")
    if s >= _key.SECP256K1_ORDER_HALF:
        s = _key.SECP256K1_ORDER - s
    return r.to_bytes(32, "little") + s.to_bytes(32, "little")


def ecdsa_verify(sig, msg, pub, context=None):
    if len(sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    if len(msg) != 32:
        raise ValueError("Message should be 32 bytes long")
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    pubkey = _pubkey_parse(pub)
    return pubkey.verify_ecdsa(ecdsa_signature_serialize_der(sig), msg)


def ecdsa_sign(msg, secret, nonce_function=None, extra_data=None, context=None):
    if len(msg) != 32:
        raise ValueError("Message should be 32 bytes long")
    if len(secret) != 32:
        raise ValueError("Secret key should be 32 bytes long")
    pk = _key.ECKey()
    pk.set(secret, False)
    sig = pk.sign_ecdsa(msg, nonce_function, extra_data)
    return ecdsa_signature_parse_der(sig)


def ec_seckey_verify(secret, context=None):
    if len(secret) != 32:
        raise ValueError("Secret should be 32 bytes long")
    pk = _key.ECKey()
    pk.set(secret, compressed=False)
    return pk.is_valid


def ec_privkey_negate(secret, context=None):
    # negate in place
    if len(secret) != 32:
        raise ValueError("Secret should be 32 bytes long")
    s = int.from_bytes(secret, "big")
    s2 = _key.SECP256K1_ORDER - s
    return s2.to_bytes(32, "big")


def ec_pubkey_negate(pubkey, context=None):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be a 64-byte structure")
    sec = ec_pubkey_serialize(pubkey)
    return ec_pubkey_parse(bytes([0x05 - sec[0]]) + sec[1:])


def ec_privkey_tweak_add(secret, tweak, context=None):
    res = ec_privkey_add(secret, tweak)
    for i in range(len(secret)):
        secret[i] = res[i]


def ec_pubkey_tweak_add(pub, tweak, context=None):
    res = ec_pubkey_add(pub, tweak)
    for i in range(len(pub)):
        pub[i] = res[i]


def ec_privkey_add(secret, tweak, context=None):
    if len(secret) != 32 or len(tweak) != 32:
        raise ValueError("Secret and tweak should both be 32 bytes long")
    s = int.from_bytes(secret, "big")
    t = int.from_bytes(tweak, "big")
    r = (s + t) % _key.SECP256K1_ORDER
    return r.to_bytes(32, "big")


def ec_pubkey_add(pub, tweak, context=None):
    if len(pub) != 64:
        raise ValueError("Public key should be 64 bytes long")
    if len(tweak) != 32:
        raise ValueError("Tweak should be 32 bytes long")
    pubkey = _pubkey_parse(pub)
    pubkey.compressed = True
    t = int.from_bytes(tweak, "big")
    Q = _key.SECP256K1.affine(
        _key.SECP256K1.mul([(_key.SECP256K1_G, t), (pubkey.p, 1)])
    )
    if Q is None:
        return None
    return Q[0].to_bytes(32, "little") + Q[1].to_bytes(32, "little")


# def ec_privkey_tweak_mul(secret, tweak, context=None):
#     if len(secret)!=32 or len(tweak)!=32:
#         raise ValueError("Secret and tweak should both be 32 bytes long")
#     s = int.from_bytes(secret, 'big')
#     t = int.from_bytes(tweak, 'big')
#     if t > _key.SECP256K1_ORDER or s > _key.SECP256K1_ORDER:
#         raise ValueError("Failed to tweak the secret")
#     r = pow(s, t, _key.SECP256K1_ORDER)
#     res = r.to_bytes(32, 'big')
#     for i in range(len(secret)):
#         secret[i] = res[i]

# def ec_pubkey_tweak_mul(pub, tweak, context=None):
#     if len(pub)!=64:
#         raise ValueError("Public key should be 64 bytes long")
#     if len(tweak)!=32:
#         raise ValueError("Tweak should be 32 bytes long")
#     if _secp.secp256k1_ec_pubkey_tweak_mul(context, pub, tweak) == 0:
#         raise ValueError("Failed to tweak the public key")

# def ec_pubkey_combine(*args, context=None):
#     pub = bytes(64)
#     pubkeys = (c_char_p * len(args))(*args)
#     r = _secp.secp256k1_ec_pubkey_combine(context, pub, pubkeys, len(args))
#     if r == 0:
#         raise ValueError("Failed to negate pubkey")
#     return pub

# schnorrsig


def xonly_pubkey_from_pubkey(pubkey, context=None):
    if len(pubkey) != 64:
        raise ValueError("Pubkey should be 64 bytes long")
    sec = ec_pubkey_serialize(pubkey)
    parity = sec[0] == 0x03
    pub = ec_pubkey_parse(b"\x02" + sec[1:33])
    return pub, parity


def schnorrsig_verify(sig, msg, pubkey, context=None):
    assert len(sig) == 64
    assert len(msg) == 32
    assert len(pubkey) == 64
    sec = ec_pubkey_serialize(pubkey)
    return _key.verify_schnorr(sec[1:33], sig, msg)


def keypair_create(secret, context=None):
    pub = ec_pubkey_create(secret)
    pub2, parity = xonly_pubkey_from_pubkey(pub)
    keypair = secret + pub
    return keypair


def schnorrsig_sign(msg, keypair, nonce_function=None, extra_data=None, context=None):
    assert len(msg) == 32
    if len(keypair) == 32:
        keypair = keypair_create(keypair, context=context)
    assert len(keypair) == 96
    return _key.sign_schnorr(keypair[:32], msg, extra_data)


# recoverable


def ecdsa_sign_recoverable(msg, secret, context=None):
    sig = ecdsa_sign(msg, secret)
    pub = ec_pubkey_create(secret)
    # Search for correct index. Not efficient but I am lazy.
    # For efficiency use c-bindings to libsecp256k1
    for i in range(4):
        if ecdsa_recover(sig + bytes([i]), msg) == pub:
            return sig + bytes([i])
    raise ValueError("Failed to sign")


def ecdsa_recoverable_signature_serialize_compact(sig, context=None):
    if len(sig) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    compact = ecdsa_signature_serialize_compact(sig[:64])
    return compact, sig[64]


def ecdsa_recoverable_signature_parse_compact(compact_sig, recid, context=None):
    if len(compact_sig) != 64:
        raise ValueError("Signature should be 64 bytes long")
    # TODO: also check r value so recid > 2 makes sense
    if recid < 0 or recid > 4:
        raise ValueError("Failed parsing compact signature")
    return ecdsa_signature_parse_compact(compact_sig) + bytes([recid])


def ecdsa_recoverable_signature_convert(sigin, context=None):
    if len(sigin) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    return sigin[:64]


def ecdsa_recover(sig, msghash, context=None):
    if len(sig) != 65:
        raise ValueError("Recoverable signature should be 65 bytes long")
    if len(msghash) != 32:
        raise ValueError("Message should be 32 bytes long")
    idx = sig[-1]
    r = int.from_bytes(sig[:32], "little")
    s = int.from_bytes(sig[32:64], "little")
    z = int.from_bytes(msghash, "big")
    # r = Rx mod N, so R can be 02x, 03x, 02(N+x), 03(N+x)
    # two latter cases only if N+x < P
    r_candidates = [
        b"\x02" + r.to_bytes(32, "big"),
        b"\x03" + r.to_bytes(32, "big"),
    ]
    if r + _key.SECP256K1_ORDER < _key.SECP256K1_FIELD_SIZE:
        r2 = r + _key.SECP256K1_ORDER
        r_candidates = r_candidates + [
            b"\x02" + r2.to_bytes(32, "big"),
            b"\x03" + r2.to_bytes(32, "big"),
        ]
    if idx >= len(r_candidates):
        raise ValueError("Failed to recover public key")
    R = _key.ECPubKey()
    R.set(r_candidates[idx])
    # s = (z + d * r)/k
    # (R*s/r - z/r*G) = P
    rinv = _key.modinv(r, _key.SECP256K1_ORDER)
    u1 = (s * rinv) % _key.SECP256K1_ORDER
    u2 = (z * rinv) % _key.SECP256K1_ORDER
    P1 = _key.SECP256K1.mul([(R.p, u1)])
    P2 = _key.SECP256K1.negate(_key.SECP256K1.mul([(_key.SECP256K1_G, u2)]))
    P = _key.SECP256K1.affine(_key.SECP256K1.add(P1, P2))
    result = P[0].to_bytes(32, "little") + P[1].to_bytes(32, "little")
    # verify signature at the end
    pubkey = _pubkey_parse(result)
    if not pubkey.is_valid:
        raise ValueError("Failed to recover public key")
    if not ecdsa_verify(sig[:64], msghash, result):
        raise ValueError("Failed to recover public key")
    return result



class UnknownDeviceError(Exception):
    pass


class SecurityStatusNotSatisfiedError(Exception):
    pass


class DenyError(Exception):
    pass


class IncorrectDataError(Exception):
    pass


class NotSupportedError(Exception):
    pass


class WrongP1P2Error(Exception):
    pass


class WrongDataLengthError(Exception):
    pass


class SwapError(Exception):
    pass


class InsNotSupportedError(Exception):
    pass


class ClaNotSupportedError(Exception):
    pass


class WrongResponseLengthError(Exception):
    pass


class BadStateError(Exception):
    pass


class SignatureFailError(Exception):
    pass


# Not really an error
class InterruptedExecution(Exception):
    pass



# Copyright (c) 2021 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
#
# Taken from https://github.com/bitcoin/bitcoin/blob/124e75a41ea0f3f0e90b63b0c41813184ddce2ab/test/functional/test_framework/ripemd160.py

"""
Pure Python RIPEMD160 implementation.

WARNING: This implementation is NOT constant-time.
Do not use without understanding the implications.
"""

# Message schedule indexes for the left path.
ML = [
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
    7, 4, 13, 1, 10, 6, 15, 3, 12, 0, 9, 5, 2, 14, 11, 8,
    3, 10, 14, 4, 9, 15, 8, 1, 2, 7, 0, 6, 13, 11, 5, 12,
    1, 9, 11, 10, 0, 8, 12, 4, 13, 3, 7, 15, 14, 5, 6, 2,
    4, 0, 5, 9, 7, 12, 2, 10, 14, 1, 3, 8, 11, 6, 15, 13
]

# Message schedule indexes for the right path.
MR = [
    5, 14, 7, 0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12,
    6, 11, 3, 7, 0, 13, 5, 10, 14, 15, 8, 12, 4, 9, 1, 2,
    15, 5, 1, 3, 7, 14, 6, 9, 11, 8, 12, 2, 10, 0, 4, 13,
    8, 6, 4, 1, 3, 11, 15, 0, 5, 12, 2, 13, 9, 7, 10, 14,
    12, 15, 10, 4, 1, 5, 8, 7, 6, 2, 13, 14, 0, 3, 9, 11
]

# Rotation counts for the left path.
RL = [
    11, 14, 15, 12, 5, 8, 7, 9, 11, 13, 14, 15, 6, 7, 9, 8,
    7, 6, 8, 13, 11, 9, 7, 15, 7, 12, 15, 9, 11, 7, 13, 12,
    11, 13, 6, 7, 14, 9, 13, 15, 14, 8, 13, 6, 5, 12, 7, 5,
    11, 12, 14, 15, 14, 15, 9, 8, 9, 14, 5, 6, 8, 6, 5, 12,
    9, 15, 5, 11, 6, 8, 13, 12, 5, 12, 13, 14, 11, 8, 5, 6
]

# Rotation counts for the right path.
RR = [
    8, 9, 9, 11, 13, 15, 15, 5, 7, 7, 8, 11, 14, 14, 12, 6,
    9, 13, 15, 7, 12, 8, 9, 11, 7, 7, 12, 7, 6, 15, 13, 11,
    9, 7, 15, 11, 8, 6, 6, 14, 12, 13, 5, 14, 13, 13, 7, 5,
    15, 5, 8, 11, 14, 14, 6, 14, 6, 9, 12, 9, 12, 5, 15, 8,
    8, 5, 12, 9, 12, 5, 14, 6, 8, 13, 6, 5, 15, 13, 11, 11
]

# K constants for the left path.
KL = [0, 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xa953fd4e]

# K constants for the right path.
KR = [0x50a28be6, 0x5c4dd124, 0x6d703ef3, 0x7a6d76e9, 0]


def fi(x, y, z, i):
    """The f1, f2, f3, f4, and f5 functions from the specification."""
    if i == 0:
        return x ^ y ^ z
    elif i == 1:
        return (x & y) | (~x & z)
    elif i == 2:
        return (x | ~y) ^ z
    elif i == 3:
        return (x & z) | (y & ~z)
    elif i == 4:
        return x ^ (y | ~z)
    else:
        assert False


def rol(x, i):
    """Rotate the bottom 32 bits of x left by i bits."""
    return ((x << i) | ((x & 0xffffffff) >> (32 - i))) & 0xffffffff


def compress(h0, h1, h2, h3, h4, block):
    """Compress state (h0, h1, h2, h3, h4) with block."""
    # Left path variables.
    al, bl, cl, dl, el = h0, h1, h2, h3, h4
    # Right path variables.
    ar, br, cr, dr, er = h0, h1, h2, h3, h4
    # Message variables.
    x = [int.from_bytes(block[4*i:4*(i+1)], 'little') for i in range(16)]

    # Iterate over the 80 rounds of the compression.
    for j in range(80):
        rnd = j >> 4
        # Perform left side of the transformation.
        al = rol(al + fi(bl, cl, dl, rnd) + x[ML[j]] + KL[rnd], RL[j]) + el
        al, bl, cl, dl, el = el, al, bl, rol(cl, 10), dl
        # Perform right side of the transformation.
        ar = rol(ar + fi(br, cr, dr, 4 - rnd) + x[MR[j]] + KR[rnd], RR[j]) + er
        ar, br, cr, dr, er = er, ar, br, rol(cr, 10), dr

    # Compose old state, left transform, and right transform into new state.
    return h1 + cl + dr, h2 + dl + er, h3 + el + ar, h4 + al + br, h0 + bl + cr


def ripemd160(data):
    """Compute the RIPEMD-160 hash of data."""
    # Initialize state.
    state = (0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0)
    # Process full 64-byte blocks in the input.
    for b in range(len(data) >> 6):
        state = compress(*state, data[64*b:64*(b+1)])
    # Construct final blocks (with padding and size).
    pad = b"\x80" + b"\x00" * ((119 - len(data)) & 63)
    fin = data[len(data) & ~63:] + pad + (8 * len(data)).to_bytes(8, 'little')
    # Process final blocks.
    for b in range(len(fin) >> 6):
        state = compress(*state, fin[64*b:64*(b+1)])
    # Produce output.
    return b"".join((h & 0xffffffff).to_bytes(4, 'little') for h in state)



"""
This module provides a compatibility layer between the python client of the Ledger Nano Bitcoin app v2 and the v1.6.5,
by translating client requests to the API of the app v1.6.5.

The bulk of the code is taken from bitcoin-core/HWI, with the necessary adaptations.
https://github.com/bitcoin-core/HWI/tree/a109bcd53d24a52e72f26af3ecbabb64b292ff0c,
"""

import struct
import re
import base64

from .client_base import PartialSignature
from .client import Client, TransportClient

from typing import List, Tuple, Optional, Union

from .common import AddressType, Chain, hash160
from .key import ExtendedKey, parse_path
from .psbt import PSBT, normalize_psbt
from .wallet import WalletPolicy

from ._script import is_p2sh, is_witness, is_p2wpkh, is_p2wsh

from .btchip.btchip import btchip
from .btchip.btchipUtils import compress_public_key
from .btchip.bitcoinTransaction import bitcoinTransaction


def get_address_type_for_policy(policy: WalletPolicy) -> AddressType:
    if policy.descriptor_template in ["pkh(@0/**)", "pkh(@0/<0;1>/*)"]:
        return AddressType.LEGACY
    elif policy.descriptor_template in ["wpkh(@0/**)", "wpkh(@0/<0:1>/*)"]:
        return AddressType.WIT
    elif policy.descriptor_template in ["sh(wpkh(@0/**))", "sh(wpkh(@0/<0;1>/*))"]:
        return AddressType.SH_WIT
    else:
        raise ValueError("Invalid or unsupported policy")


# minimal checking of string keypath
# taken from HWI
def check_keypath(key_path: str) -> bool:
    parts = re.split("/", key_path)
    if parts[0] != "m":
        return False
    # strip hardening chars
    for index in parts[1:]:
        index_int = re.sub('[hH\']', '', index)
        if not index_int.isdigit():
            return False
        if int(index_int) > 0x80000000:
            return False
    return True


class DongleAdaptor:
    # TODO: type for comm_client
    def __init__(self, comm_client):
        self.comm_client = comm_client

    def exchange(self, apdu: Union[bytes, bytearray]) -> bytearray:
        cla = apdu[0]
        ins = apdu[1]
        p1 = apdu[2]
        p2 = apdu[3]
        lc = apdu[4]
        data = apdu[5:]
        assert len(data) == lc
        return bytearray(self.comm_client.apdu_exchange(cla, ins, data, p1, p2))

class LegacyClient(Client):
    """Wrapper for Ledger Bitcoin app before version 2.0.0."""

    def __init__(self, comm_client: TransportClient, chain: Chain = Chain.MAIN, debug: bool = False):
        super().__init__(comm_client, chain, debug)

        self.app = btchip(DongleAdaptor(comm_client))

        if self.app.getAppName() not in ["Bitcoin", "Bitcoin Legacy", "Bitcoin Test", "Bitcoin Test Legacy", "app"]:
            raise ValueError("Ledger is not in either the Bitcoin or Bitcoin Testnet app")

    def get_extended_pubkey(self, path: str, display: bool = False) -> str:
        # mostly taken from HWI

        path = path[2:]
        path = path.replace('h', '\'')
        path = path.replace('H', '\'')

        # This call returns raw uncompressed pubkey, chaincode
        pubkey = self.app.getWalletPublicKey(path, display)
        int_path = parse_path(path)
        if len(path) > 0:
            parent_path = ""
            for ind in path.split("/")[:-1]:
                parent_path += ind + "/"
            parent_path = parent_path[:-1]

            # Get parent key fingerprint
            parent = self.app.getWalletPublicKey(parent_path)
            fpr = hash160(compress_public_key(parent["publicKey"]))[:4]

            child = int_path[-1]
        # Special case for m
        else:
            child = 0
            fpr = b"\x00\x00\x00\x00"

        xpub = ExtendedKey(
            version=ExtendedKey.MAINNET_PUBLIC if self.chain == Chain.MAIN else ExtendedKey.TESTNET_PUBLIC,
            depth=len(path.split("/")) if len(path) > 0 else 0,
            parent_fingerprint=fpr,
            child_num=child,
            chaincode=pubkey["chainCode"],
            privkey=None,
            pubkey=compress_public_key(pubkey["publicKey"]),
        )
        return xpub.to_string()

    def register_wallet(self, wallet: WalletPolicy) -> Tuple[bytes, bytes]:
        raise NotImplementedError  # legacy app does not have this functionality

    def get_wallet_address(
        self,
        wallet: WalletPolicy,
        wallet_hmac: Optional[bytes],
        change: int,
        address_index: int,
        display: bool,
    ) -> str:
        # TODO: check keypath

        if wallet_hmac is not None or wallet.n_keys != 1:
            raise NotImplementedError("Policy wallets are only supported from version 2.0.0. Please update your Ledger hardware wallet")

        if not isinstance(wallet, WalletPolicy):
            raise ValueError("Invalid wallet policy type, it must be WalletPolicy")

        key_info = wallet.keys_info[0]
        try:
            first_slash_pos = key_info.index("/")
            key_origin_end = key_info.index("]")
        except ValueError:
            raise ValueError("Could not extract key origin information")

        if key_info[0] != '[':
            raise ValueError("Key must have key origin information")

        key_origin_path = key_info[first_slash_pos + 1: key_origin_end]

        addr_type = get_address_type_for_policy(wallet)

        p2sh_p2wpkh = addr_type == AddressType.SH_WIT
        bech32 = addr_type == AddressType.WIT
        output = self.app.getWalletPublicKey(f"{key_origin_path}/{change}/{address_index}", display, p2sh_p2wpkh or bech32, bech32)
        assert isinstance(output["address"], str)
        return output['address'][12:-2]  # HACK: A bug in getWalletPublicKey results in the address being returned as the string "bytearray(b'<address>')". This extracts the actual address to work around this.

    def sign_psbt(self, psbt: Union[PSBT, bytes, str], wallet: WalletPolicy, wallet_hmac: Optional[bytes]) -> List[Tuple[int, PartialSignature]]:
        if wallet_hmac is not None or wallet.n_keys != 1:
            raise NotImplementedError("Policy wallets are only supported from version 2.0.0. Please update your Ledger hardware wallet")

        if not isinstance(wallet, WalletPolicy):
            raise ValueError("Invalid wallet policy type, it must be WalletPolicy")

        if wallet.descriptor_template not in ["pkh(@0/**)", "pkh(@0/<0;1>/*)", "wpkh(@0/**)", "wpkh(@0/<0;1>/*)", "sh(wpkh(@0/**))", "sh(wpkh(@0/<0;1>/*))"]:
            raise NotImplementedError("Unsupported policy")

        psbt = normalize_psbt(psbt)

        # the rest of the code is basically the HWI code, and it ignores wallet

        tx = psbt

        #c_tx = tx.get_unsigned_tx()
        c_tx = tx.tx
        tx_bytes = c_tx.serialize_with_witness()

        # Master key fingerprint
        master_fpr = hash160(compress_public_key(self.app.getWalletPublicKey('')["publicKey"]))[:4]
        # An entry per input, each with 0 to many keys to sign with
        all_signature_attempts: List[List[Tuple[str, bytes]]] = [[]] * len(c_tx.vin)

        # Get the app version to determine whether to use Trusted Input for segwit
        version = self.app.getFirmwareVersion()
        use_trusted_segwit = (version['major_version'] == 1 and version['minor_version'] >= 4) or version['major_version'] > 1

        # NOTE: We only support signing Segwit inputs, where we can skip over non-segwit
        # inputs, or non-segwit inputs, where *all* inputs are non-segwit. This is due
        # to Ledger's mutually exclusive signing steps for each type.
        segwit_inputs = []
        # Legacy style inputs
        legacy_inputs = []

        has_segwit = False
        has_legacy = False

        script_codes: List[bytes] = [b""] * len(c_tx.vin)

        # Detect changepath, (p2sh-)p2(w)pkh only
        change_path = ''
        for txout, i_num in zip(c_tx.vout, range(len(c_tx.vout))):
            # Find which wallet key could be change based on hdsplit: m/.../1/k
            # Wallets shouldn't be sending to change address as user action
            # otherwise this will get confused
            for pubkey, origin in tx.outputs[i_num].hd_keypaths.items():
                if origin.fingerprint == master_fpr and len(origin.path) > 1 and origin.path[-2] == 1:
                    # For possible matches, check if pubkey matches possible template
                    if hash160(pubkey) in txout.scriptPubKey or hash160(bytearray.fromhex("0014") + hash160(pubkey)) in txout.scriptPubKey:
                        change_path = ''
                        for index in origin.path:
                            change_path += str(index) + "/"
                        change_path = change_path[:-1]

        for txin, psbt_in, i_num in zip(c_tx.vin, tx.inputs, range(len(c_tx.vin))):

            seq_hex = txin.nSequence.to_bytes(4, byteorder="little").hex()

            scriptcode = b""
            utxo = None
            if psbt_in.witness_utxo:
                utxo = psbt_in.witness_utxo
            if psbt_in.non_witness_utxo:
                if txin.prevout.hash != psbt_in.non_witness_utxo.sha256:
                    raise ValueError('Input {} has a non_witness_utxo with the wrong hash'.format(i_num))
                utxo = psbt_in.non_witness_utxo.vout[txin.prevout.n]
            if utxo is None:
                raise Exception("PSBT is missing input utxo information, cannot sign")
            scriptcode = utxo.scriptPubKey

            if is_p2sh(scriptcode):
                if len(psbt_in.redeem_script) == 0:
                    continue
                scriptcode = psbt_in.redeem_script

            is_wit, _, _ = is_witness(scriptcode)

            segwit_inputs.append({"value": txin.prevout.serialize() + struct.pack("<Q", utxo.nValue), "witness": True, "sequence": seq_hex})
            if is_wit:
                if is_p2wsh(scriptcode):
                    if len(psbt_in.witness_script) == 0:
                        continue
                    scriptcode = psbt_in.witness_script
                elif is_p2wpkh(scriptcode):
                    _, _, wit_prog = is_witness(scriptcode)
                    scriptcode = b"\x76\xa9\x14" + wit_prog + b"\x88\xac"
                else:
                    continue
                has_segwit = True
            else:
                # We only need legacy inputs in the case where all inputs are legacy, we check
                # later
                assert psbt_in.non_witness_utxo is not None
                ledger_prevtx = bitcoinTransaction(psbt_in.non_witness_utxo.serialize())
                legacy_inputs.append(self.app.getTrustedInput(ledger_prevtx, txin.prevout.n))
                legacy_inputs[-1]["sequence"] = seq_hex
                has_legacy = True

            if psbt_in.non_witness_utxo and use_trusted_segwit:
                ledger_prevtx = bitcoinTransaction(psbt_in.non_witness_utxo.serialize())
                segwit_inputs[-1].update(self.app.getTrustedInput(ledger_prevtx, txin.prevout.n))

            pubkeys = []
            signature_attempts = []

            # Save scriptcode for later signing
            script_codes[i_num] = scriptcode

            # Find which pubkeys could sign this input (should be all?)
            for pubkey in psbt_in.hd_keypaths.keys():
                if hash160(pubkey) in scriptcode or pubkey in scriptcode:
                    pubkeys.append(pubkey)

            # Figure out which keys in inputs are from our wallet
            for pubkey in pubkeys:
                keypath = psbt_in.hd_keypaths[pubkey]
                if master_fpr == keypath.fingerprint:
                    # Add the keypath strings
                    keypath_str = keypath.get_derivation_path()[2:] # Drop the leading m/
                    signature_attempts.append((keypath_str, pubkey))

            all_signature_attempts[i_num] = signature_attempts

        result: List[Tuple(int, PartialSignature)] = []

        # Sign any segwit inputs
        if has_segwit:
            # Process them up front with all scriptcodes blank
            blank_script_code = bytearray()
            for i in range(len(segwit_inputs)):
                self.app.startUntrustedTransaction(i == 0, i, segwit_inputs, script_codes[i] if use_trusted_segwit else blank_script_code, c_tx.nVersion)

            # Number of unused fields for Nano S, only changepath and transaction in bytes req
            self.app.finalizeInput(b"DUMMY", -1, -1, change_path, tx_bytes)

            # For each input we control do segwit signature
            for i in range(len(segwit_inputs)):
                for signature_attempt in all_signature_attempts[i]:
                    self.app.startUntrustedTransaction(False, 0, [segwit_inputs[i]], script_codes[i], c_tx.nVersion)

                    # tx.inputs[i].partial_sigs[signature_attempt[1]] = self.app.untrustedHashSign(signature_attempt[0], "", c_tx.nLockTime, 0x01)

                    partial_sig = PartialSignature(
                        signature=self.app.untrustedHashSign(signature_attempt[0], "", c_tx.nLockTime, 0x01),
                        pubkey=signature_attempt[1]
                    )
                    result.append((i, partial_sig))
        elif has_legacy:
            first_input = True
            # Legacy signing if all inputs are legacy
            for i in range(len(legacy_inputs)):
                for signature_attempt in all_signature_attempts[i]:
                    assert (tx.inputs[i].non_witness_utxo is not None)
                    self.app.startUntrustedTransaction(first_input, i, legacy_inputs, script_codes[i], c_tx.nVersion)
                    self.app.finalizeInput(b"DUMMY", -1, -1, change_path, tx_bytes)

                    #tx.inputs[i].partial_sigs[signature_attempt[1]] = self.app.untrustedHashSign(signature_attempt[0], "", c_tx.nLockTime, 0x01)

                    partial_sig = PartialSignature(
                        signature=self.app.untrustedHashSign(signature_attempt[0], "", c_tx.nLockTime, 0x01),
                        pubkey=signature_attempt[1]
                    )
                    result.append((i, partial_sig))

                    first_input = False

        # Send list of input signatures
        return result

    def get_master_fingerprint(self) -> bytes:
        master_pubkey = self.app.getWalletPublicKey("")
        return hash160(compress_public_key(master_pubkey["publicKey"]))[:4]

    def sign_message(self, message: Union[str, bytes], keypath: str) -> str:
        # copied verbatim from HWI

        if not check_keypath(keypath):
            raise ValueError("Invalid keypath")
        if isinstance(message, str):
            message = bytearray(message, 'utf-8')
        else:
            message = bytearray(message)
        keypath = keypath[2:]
        # First display on screen what address you're signing for
        self.app.getWalletPublicKey(keypath, True)
        self.app.signMessagePrepare(keypath, message)
        signature = self.app.signMessageSign()

        # Make signature into standard bitcoin format
        rLength = signature[3]
        r = int.from_bytes(signature[4: 4 + rLength], byteorder="big", signed=True)
        s = int.from_bytes(signature[4 + rLength + 2:], byteorder="big", signed=True)

        sig = bytearray(chr(27 + 4 + (signature[0] & 0x01)), 'utf8') + r.to_bytes(32, byteorder="big", signed=False) + s.to_bytes(32, byteorder="big", signed=False)

        return base64.b64encode(sig).decode('utf-8')



from pathlib import Path

from bitcoin_client.ledger_bitcoin import Client, WalletPolicy, PartialSignature

from bitcoin_client.ledger_bitcoin.psbt import PSBT

from test_utils import has_automation

tests_root: Path = Path(__file__).parent


CURRENCY_TICKER = "TEST"


def open_psbt_from_file(filename: str) -> PSBT:
    raw_psbt_base64 = open(filename, "r").read()

    psbt = PSBT()
    psbt.deserialize(raw_psbt_base64)
    return psbt


@has_automation("automations/sign_with_wallet_accept.json")
def test_sign_psbt_singlesig_pkh_1to1(client: Client):

    # PSBT for a legacy 1-input 1-output spend (no change address)
    psbt = open_psbt_from_file(f"{tests_root}/psbt/singlesig/pkh-1to1.psbt")

    wallet = WalletPolicy(
        "",
        "pkh(@0/**)",
        [
            "[f5acc2fd/44'/1'/0']tpubDCwYjpDhUdPGP5rS3wgNg13mTrrjBuG8V9VpWbyptX6TRPbNoZVXsoVUSkCjmQ8jJycjuDKBb9eataSymXakTTaGifxR6kmVsfFehH1ZgJT"
        ],
    )

    # expected sigs:
    # #0:
    #  "pubkey" : "02ee8608207e21028426f69e76447d7e3d5e077049f5e683c3136c2314762a4718",
    #  "signature" : "3045022100e55b3ca788721aae8def2eadff710e524ffe8c9dec1764fdaa89584f9726e196022012a30fbcf9e1a24df31a1010356b794ab8de438b4250684757ed5772402540f401"
    result = client.sign_psbt(psbt, wallet, None)

    assert result == [(
        0,
        PartialSignature(
            pubkey=bytes.fromhex("02ee8608207e21028426f69e76447d7e3d5e077049f5e683c3136c2314762a4718"),
            signature=bytes.fromhex(
                "3045022100e55b3ca788721aae8def2eadff710e524ffe8c9dec1764fdaa89584f9726e196022012a30fbcf9e1a24df31a1010356b794ab8de438b4250684757ed5772402540f401"
            )
        )
    )]


@has_automation("automations/sign_with_wallet_accept.json")
def test_sign_psbt_singlesig_sh_wpkh_1to2(client: Client):

    # PSBT for a wrapped segwit 1-input 2-output spend (1 change address)
    psbt = open_psbt_from_file(f"{tests_root}/psbt/singlesig/sh-wpkh-1to2.psbt")

    wallet = WalletPolicy(
        "",
        "sh(wpkh(@0/**))",
        [
            "[f5acc2fd/49'/1'/0']tpubDC871vGLAiKPcwAw22EjhKVLk5L98UGXBEcGR8gpcigLQVDDfgcYW24QBEyTHTSFEjgJgbaHU8CdRi9vmG4cPm1kPLmZhJEP17FMBdNheh3"
        ],
    )

    # expected sigs:
    # #0:
    #  "pubkey" : "024ba3b77d933de9fa3f9583348c40f3caaf2effad5b6e244ece8abbfcc7244f67",
    #  "signature" : "30440220720722b08489c2a50d10edea8e21880086c8e8f22889a16815e306daeea4665b02203fcf453fa490b76cf4f929714065fc90a519b7b97ab18914f9451b5a4b45241201"
    result = client.sign_psbt(psbt, wallet, None)

    assert len(result) == 1

    assert result == [(
        0,
        PartialSignature(
            pubkey=bytes.fromhex("024ba3b77d933de9fa3f9583348c40f3caaf2effad5b6e244ece8abbfcc7244f67"),
            signature=bytes.fromhex(
                "30440220720722b08489c2a50d10edea8e21880086c8e8f22889a16815e306daeea4665b02203fcf453fa490b76cf4f929714065fc90a519b7b97ab18914f9451b5a4b45241201"
            )
        )
    )]


@has_automation("automations/sign_with_wallet_accept.json")
def test_sign_psbt_singlesig_wpkh_1to2(client: Client):

    # PSBT for a legacy 1-input 2-output spend (1 change address)
    psbt = open_psbt_from_file(f"{tests_root}/psbt/singlesig/wpkh-1to2.psbt")

    wallet = WalletPolicy(
        "",
        "wpkh(@0/**)",
        [
            "[f5acc2fd/84'/1'/0']tpubDCtKfsNyRhULjZ9XMS4VKKtVcPdVDi8MKUbcSD9MJDyjRu1A2ND5MiipozyyspBT9bg8upEp7a8EAgFxNxXn1d7QkdbL52Ty5jiSLcxPt1P"
        ],
    )

    result = client.sign_psbt(psbt, wallet, None)

    # expected sigs
    # #0:
    #   "pubkey" : "03ee2c3d98eb1f93c0a1aa8e5a4009b70eb7b44ead15f1666f136b012ad58d3068",
    #   "signature" : "3045022100ab44f34dd7e87c9054591297a101e8500a0641d1d591878d0d23cf8096fa79e802205d12d1062d925e27b57bdcf994ecf332ad0a8e67b8fe407bab2101255da632aa01"

    assert result == [(
        0,
        PartialSignature(
            pubkey=bytes.fromhex("03ee2c3d98eb1f93c0a1aa8e5a4009b70eb7b44ead15f1666f136b012ad58d3068"),
            signature=bytes.fromhex(
                "3045022100ab44f34dd7e87c9054591297a101e8500a0641d1d591878d0d23cf8096fa79e802205d12d1062d925e27b57bdcf994ecf332ad0a8e67b8fe407bab2101255da632aa01"
            )
        )
    )]


@has_automation("automations/sign_with_wallet_accept.json")
def test_sign_psbt_singlesig_wpkh_2to2(client: Client):
    # PSBT for a legacy 2-input 2-output spend (1 change address)

    psbt = open_psbt_from_file(f"{tests_root}/psbt/singlesig/wpkh-2to2.psbt")

    wallet = WalletPolicy(
        "",
        "wpkh(@0/**)",
        [
            "[f5acc2fd/84'/1'/0']tpubDCtKfsNyRhULjZ9XMS4VKKtVcPdVDi8MKUbcSD9MJDyjRu1A2ND5MiipozyyspBT9bg8upEp7a8EAgFxNxXn1d7QkdbL52Ty5jiSLcxPt1P"
        ],
    )

    result = client.sign_psbt(psbt, wallet, None)

    assert result == [(
        0,
        PartialSignature(
            pubkey=bytes.fromhex("03455ee7cedc97b0ba435b80066fc92c963a34c600317981d135330c4ee43ac7a3"),
            signature=bytes.fromhex(
                "304402206b3e877655f08c6e7b1b74d6d893a82cdf799f68a5ae7cecae63a71b0339e5ce022019b94aa3fb6635956e109f3d89c996b1bfbbaf3c619134b5a302badfaf52180e01"
            )
        )
    ), (
        1,
        PartialSignature(
            pubkey=bytes.fromhex("0271b5b779ad870838587797bcf6f0c7aec5abe76a709d724f48d2e26cf874f0a0"),
            signature=bytes.fromhex(
                "3045022100e2e98e4f8c70274f10145c89a5d86e216d0376bdf9f42f829e4315ea67d79d210220743589fd4f55e540540a976a5af58acd610fa5e188a5096dfe7d36baf3afb94001"
            )
        )
    )]



# Original version: https://raw.githubusercontent.com/bitcoin-core/HWI/3fe369d0379212fae1c72729a179d133b0adc872/hwilib/_serialize.py
#
#!/usr/bin/env python3
# Copyright (c) 2010 ArtForz -- public domain half-a-node
# Copyright (c) 2012 Jeff Garzik
# Copyright (c) 2010-2016 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""
Bitcoin Object Python Serializations
************************************

Modified from the test/test_framework/mininode.py file from the
Bitcoin repository
"""

import struct

from typing import (
    List,
    Sequence,
    TypeVar,
    Callable,
)
from typing_extensions import Protocol

class Readable(Protocol):
    def read(self, n: int = -1) -> bytes:
        ...

class Deserializable(Protocol):
    def deserialize(self, f: Readable) -> None:
        ...

class Serializable(Protocol):
    def serialize(self) -> bytes:
        ...


# Serialization/deserialization tools
def ser_compact_size(size: int) -> bytes:
    """
    Serialize an integer using Bitcoin's compact size unsigned integer serialization.

    :param size: The int to serialize
    :returns: The int serialized as a compact size unsigned integer
    """
    r = b""
    if size < 253:
        r = struct.pack("B", size)
    elif size < 0x10000:
        r = struct.pack("<BH", 253, size)
    elif size < 0x100000000:
        r = struct.pack("<BI", 254, size)
    else:
        r = struct.pack("<BQ", 255, size)
    return r

def deser_compact_size(f: Readable) -> int:
    """
    Deserialize a compact size unsigned integer from the beginning of the byte stream.

    :param f: The byte stream
    :returns: The integer that was serialized
    """
    nit: int = struct.unpack("<B", f.read(1))[0]
    if nit == 253:
        nit = struct.unpack("<H", f.read(2))[0]
    elif nit == 254:
        nit = struct.unpack("<I", f.read(4))[0]
    elif nit == 255:
        nit = struct.unpack("<Q", f.read(8))[0]
    return nit

def deser_string(f: Readable) -> bytes:
    """
    Deserialize a variable length byte string serialized with Bitcoin's variable length string serialization from a byte stream.

    :param f: The byte stream
    :returns: The byte string that was serialized
    """
    nit = deser_compact_size(f)
    return f.read(nit)

def ser_string(s: bytes) -> bytes:
    """
    Serialize a byte string with Bitcoin's variable length string serialization.

    :param s: The byte string to be serialized
    :returns: The serialized byte string
    """
    return ser_compact_size(len(s)) + s

def deser_uint256(f: Readable) -> int:
    """
    Deserialize a 256 bit integer serialized with Bitcoin's 256 bit integer serialization from a byte stream.

    :param f: The byte stream.
    :returns: The integer that was serialized
    """
    r = 0
    for i in range(8):
        t = struct.unpack("<I", f.read(4))[0]
        r += t << (i * 32)
    return r


def ser_uint256(u: int) -> bytes:
    """
    Serialize a 256 bit integer with Bitcoin's 256 bit integer serialization.

    :param u: The integer to serialize
    :returns: The serialized 256 bit integer
    """
    rs = b""
    for _ in range(8):
        rs += struct.pack("<I", u & 0xFFFFFFFF)
        u >>= 32
    return rs


def uint256_from_str(s: bytes) -> int:
    """
    Deserialize a 256 bit integer serialized with Bitcoin's 256 bit integer serialization from a byte string.

    :param s: The byte string
    :returns: The integer that was serialized
    """
    r = 0
    t = struct.unpack("<IIIIIIII", s[:32])
    for i in range(8):
        r += t[i] << (i * 32)
    return r

D = TypeVar("D", bound=Deserializable)

def deser_vector(f: Readable, c: Callable[[], D]) -> List[D]:
    """
    Deserialize a vector of objects with Bitcoin's object vector serialization from a byte stream.

    :param f: The byte stream
    :param c: The class of object to deserialize for each object in the vector
    :returns: A list of objects that were serialized
    """
    nit = deser_compact_size(f)
    r = []
    for _ in range(nit):
        t = c()
        t.deserialize(f)
        r.append(t)
    return r


def ser_vector(v: Sequence[Serializable]) -> bytes:
    """
    Serialize a vector of objects with Bitcoin's object vector serialzation.

    :param v: The list of objects to serialize
    :returns: The serialized objects
    """
    r = ser_compact_size(len(v))
    for i in v:
        r += i.serialize()
    return r


def deser_string_vector(f: Readable) -> List[bytes]:
    """
    Deserialize a vector of byte strings from a byte stream.

    :param f: The byte stream
    :returns: The list of byte strings that were serialized
    """
    nit = deser_compact_size(f)
    r = []
    for _ in range(nit):
        t = deser_string(f)
        r.append(t)
    return r


def ser_string_vector(v: List[bytes]) -> bytes:
    """
    Serialize a list of byte strings as a vector of byte strings.

    :param v: The list of byte strings to serialize
    :returns: The serialized list of byte strings
    """
    r = ser_compact_size(len(v))
    for sv in v:
        r += ser_string(sv)
    return r

def ser_sig_der(r: bytes, s: bytes) -> bytes:
    """
    Serialize the ``r`` and ``s`` values of an ECDSA signature using DER.

    :param r: The ``r`` value bytes
    :param s: The ``s`` value bytes
    :returns: The DER encoded signature
    """
    sig = b"\x30"

    # Make r and s as short as possible
    ri = 0
    for b in r:
        if b == 0:
            ri += 1
        else:
            break
    r = r[ri:]
    si = 0
    for b in s:
        if b == 0:
            si += 1
        else:
            break
    s = s[si:]

    # Make positive of neg
    first = r[0]
    if first & (1 << 7) != 0:
        r = b"\x00" + r
    first = s[0]
    if first & (1 << 7) != 0:
        s = b"\x00" + s

    # Write total length
    total_len = len(r) + len(s) + 4
    sig += struct.pack("B", total_len)

    # write r
    sig += b"\x02"
    sig += struct.pack("B", len(r))
    sig += r

    # write s
    sig += b"\x02"
    sig += struct.pack("B", len(s))
    sig += s

    sig += b"\x01"
    return sig

def ser_sig_compact(r: bytes, s: bytes, recid: bytes) -> bytes:
    """
    Serialize the ``r`` and ``s`` values of an ECDSA signature using the compact signature serialization scheme.

    :param r: The ``r`` value bytes
    :param s: The ``s`` value bytes
    :returns: The compact signature
    """
    rec = struct.unpack("B", recid)[0]
    prefix = struct.pack("B", 27 + 4 + rec)

    sig = b""
    sig += prefix
    sig += r + s

    return sig



#!/usr/bin/env python3
# Copyright (c) 2010 ArtForz -- public domain half-a-node
# Copyright (c) 2012 Jeff Garzik
# Copyright (c) 2010-2016 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Bitcoin Object Python Serializations

Modified from the test/test_framework/mininode.py file from the
Bitcoin repository

CTransaction,CTxIn, CTxOut, etc....:
    data structures that should map to corresponding structures in
    bitcoin/primitives for transactions only
"""

import copy
import struct

from .common import (
    hash256,
)
from ._script import (
    is_opreturn,
    is_p2sh,
    is_p2pkh,
    is_p2pk,
    is_witness,
    is_p2wsh,
)
from ._serialize import (
    deser_uint256,
    deser_string,
    deser_string_vector,
    deser_vector,
    Readable,
    ser_uint256,
    ser_string,
    ser_string_vector,
    ser_vector,
    uint256_from_str,
)

from typing import (
    List,
    Optional,
    Tuple,
)

# Objects that map to bitcoind objects, which can be serialized/deserialized

MSG_WITNESS_FLAG = 1 << 30

class COutPoint(object):
    def __init__(self, hash: int = 0, n: int = 0xffffffff):
        self.hash = hash
        self.n = n

    def deserialize(self, f: Readable) -> None:
        self.hash = deser_uint256(f)
        self.n = struct.unpack("<I", f.read(4))[0]

    def serialize(self) -> bytes:
        r = b""
        r += ser_uint256(self.hash)
        r += struct.pack("<I", self.n)
        return r

    def __repr__(self) -> str:
        return "COutPoint(hash=%064x n=%i)" % (self.hash, self.n)


class CTxIn(object):
    def __init__(
        self,
        outpoint: Optional[COutPoint] = None,
        scriptSig: bytes = b"",
        nSequence: int = 0,
    ):
        if outpoint is None:
            self.prevout = COutPoint()
        else:
            self.prevout = outpoint
        self.scriptSig = scriptSig
        self.nSequence = nSequence

    def deserialize(self, f: Readable) -> None:
        self.prevout = COutPoint()
        self.prevout.deserialize(f)
        self.scriptSig = deser_string(f)
        self.nSequence = struct.unpack("<I", f.read(4))[0]

    def serialize(self) -> bytes:
        r = b""
        r += self.prevout.serialize()
        r += ser_string(self.scriptSig)
        r += struct.pack("<I", self.nSequence)
        return r

    def __repr__(self) -> str:
        return "CTxIn(prevout=%s scriptSig=%s nSequence=%i)" \
            % (repr(self.prevout), self.scriptSig.hex(),
               self.nSequence)


class CTxOut(object):
    def __init__(self, nValue: int = 0, scriptPubKey: bytes = b""):
        self.nValue = nValue
        self.scriptPubKey = scriptPubKey

    def deserialize(self, f: Readable) -> None:
        self.nValue = struct.unpack("<q", f.read(8))[0]
        self.scriptPubKey = deser_string(f)

    def serialize(self) -> bytes:
        r = b""
        r += struct.pack("<q", self.nValue)
        r += ser_string(self.scriptPubKey)
        return r

    def is_opreturn(self) -> bool:
        return is_opreturn(self.scriptPubKey)

    def is_p2sh(self) -> bool:
        return is_p2sh(self.scriptPubKey)

    def is_p2wsh(self) -> bool:
        return is_p2wsh(self.scriptPubKey)

    def is_p2pkh(self) -> bool:
        return is_p2pkh(self.scriptPubKey)

    def is_p2pk(self) -> bool:
        return is_p2pk(self.scriptPubKey)

    def is_witness(self) -> Tuple[bool, int, bytes]:
        return is_witness(self.scriptPubKey)

    def __repr__(self) -> str:
        return "CTxOut(nValue=%i.%08i scriptPubKey=%s)" \
            % (self.nValue // 100_000_000, self.nValue % 100_000_000, self.scriptPubKey.hex())


class CScriptWitness(object):
    def __init__(self) -> None:
        # stack is a vector of strings
        self.stack: List[bytes] = []

    def __repr__(self) -> str:
        return "CScriptWitness(%s)" % \
               (",".join([x.hex() for x in self.stack]))

    def is_null(self) -> bool:
        if self.stack:
            return False
        return True


class CTxInWitness(object):
    def __init__(self) -> None:
        self.scriptWitness = CScriptWitness()

    def deserialize(self, f: Readable) -> None:
        self.scriptWitness.stack = deser_string_vector(f)

    def serialize(self) -> bytes:
        return ser_string_vector(self.scriptWitness.stack)

    def __repr__(self) -> str:
        return repr(self.scriptWitness)

    def is_null(self) -> bool:
        return self.scriptWitness.is_null()


class CTxWitness(object):
    def __init__(self) -> None:
        self.vtxinwit: List[CTxInWitness] = []

    def deserialize(self, f: Readable) -> None:
        for i in range(len(self.vtxinwit)):
            self.vtxinwit[i].deserialize(f)

    def serialize(self) -> bytes:
        r = b""
        # This is different than the usual vector serialization --
        # we omit the length of the vector, which is required to be
        # the same length as the transaction's vin vector.
        for x in self.vtxinwit:
            r += x.serialize()
        return r

    def __repr__(self) -> str:
        return "CTxWitness(%s)" % \
               (';'.join([repr(x) for x in self.vtxinwit]))

    def is_null(self) -> bool:
        for x in self.vtxinwit:
            if not x.is_null():
                return False
        return True


class CTransaction(object):
    def __init__(self, tx: Optional['CTransaction'] = None) -> None:
        if tx is None:
            self.nVersion = 1
            self.vin: List[CTxIn] = []
            self.vout: List[CTxOut] = []
            self.wit = CTxWitness()
            self.nLockTime = 0
            self.sha256: Optional[int] = None
            self.hash: Optional[bytes] = None
        else:
            self.nVersion = tx.nVersion
            self.vin = copy.deepcopy(tx.vin)
            self.vout = copy.deepcopy(tx.vout)
            self.nLockTime = tx.nLockTime
            self.sha256 = tx.sha256
            self.hash = tx.hash
            self.wit = copy.deepcopy(tx.wit)

    def deserialize(self, f: Readable) -> None:
        self.nVersion = struct.unpack("<i", f.read(4))[0]
        self.vin = deser_vector(f, CTxIn)
        flags = 0
        if len(self.vin) == 0:
            flags = struct.unpack("<B", f.read(1))[0]
            # Not sure why flags can't be zero, but this
            # matches the implementation in bitcoind
            if (flags != 0):
                self.vin = deser_vector(f, CTxIn)
                self.vout = deser_vector(f, CTxOut)
        else:
            self.vout = deser_vector(f, CTxOut)
        if flags != 0:
            self.wit.vtxinwit = [CTxInWitness() for i in range(len(self.vin))]
            self.wit.deserialize(f)
        self.nLockTime = struct.unpack("<I", f.read(4))[0]
        self.sha256 = None
        self.hash = None

    def serialize_without_witness(self) -> bytes:
        r = b""
        r += struct.pack("<i", self.nVersion)
        r += ser_vector(self.vin)
        r += ser_vector(self.vout)
        r += struct.pack("<I", self.nLockTime)
        return r

    # Only serialize with witness when explicitly called for
    def serialize_with_witness(self) -> bytes:
        flags = 0
        if not self.wit.is_null():
            flags |= 1
        r = b""
        r += struct.pack("<i", self.nVersion)
        if flags:
            r += ser_vector([])
            r += struct.pack("<B", flags)
        r += ser_vector(self.vin)
        r += ser_vector(self.vout)
        if flags & 1:
            if (len(self.wit.vtxinwit) != len(self.vin)):
                # vtxinwit must have the same length as vin
                self.wit.vtxinwit = self.wit.vtxinwit[:len(self.vin)]
                for _ in range(len(self.wit.vtxinwit), len(self.vin)):
                    self.wit.vtxinwit.append(CTxInWitness())
            r += self.wit.serialize()
        r += struct.pack("<I", self.nLockTime)
        return r

    # Regular serialization is without witness -- must explicitly
    # call serialize_with_witness to include witness data.
    def serialize(self) -> bytes:
        return self.serialize_without_witness()

    # Recalculate the txid (transaction hash without witness)
    def rehash(self) -> None:
        self.sha256 = None
        self.calc_sha256()

    # We will only cache the serialization without witness in
    # self.sha256 and self.hash -- those are expected to be the txid.
    def calc_sha256(self, with_witness: bool = False) -> Optional[int]:
        if with_witness:
            # Don't cache the result, just return it
            return uint256_from_str(hash256(self.serialize_with_witness()))

        if self.sha256 is None:
            self.sha256 = uint256_from_str(hash256(self.serialize_without_witness()))
        self.hash = hash256(self.serialize())
        return None

    def is_null(self) -> bool:
        return len(self.vin) == 0 and len(self.vout) == 0

    def __repr__(self) -> str:
        return "CTransaction(nVersion=%i vin=%s vout=%s wit=%s nLockTime=%i)" \
            % (self.nVersion, repr(self.vin), repr(self.vout), repr(self.wit), self.nLockTime)



from .errors import DescriptorError


def polymod(c: int, val: int) -> int:
    c0 = c >> 35
    c = ((c & 0x7FFFFFFFF) << 5) ^ val
    if c0 & 1:
        c ^= 0xF5DEE51989
    if c0 & 2:
        c ^= 0xA9FDCA3312
    if c0 & 4:
        c ^= 0x1BAB10E32D
    if c0 & 8:
        c ^= 0x3706B1677A
    if c0 & 16:
        c ^= 0x644D626FFD
    return c


def checksum(desc: str) -> str:
    """Calculate checksum of desciptor string"""
    INPUT_CHARSET = (
        "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVW"
        'XYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#"\\ '
    )
    CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"

    c = 1
    cls = 0
    clscount = 0
    for ch in desc:
        pos = INPUT_CHARSET.find(ch)
        if pos == -1:
            raise DescriptorError("Invalid character '%s' in the input string" % ch)
        c = polymod(c, pos & 31)
        cls = cls * 3 + (pos >> 5)
        clscount += 1
        if clscount == 3:
            c = polymod(c, cls)
            cls = 0
            clscount = 0
    if clscount > 0:
        c = polymod(c, cls)
    for j in range(0, 8):
        c = polymod(c, 0)
    c ^= 1

    ret = [CHECKSUM_CHARSET[(c >> (5 * (7 - j))) & 31] for j in range(0, 8)]
    return "".join(ret)


def add_checksum(desc: str) -> str:
    """Add checksum to descriptor string"""
    if "#" in desc:
        desc = desc.split("#")[0]
    return desc + "#" + checksum(desc)



from dataclasses import dataclass
from typing import List, Tuple, Optional, Union, Literal
from io import BytesIO

from ledgercomm.interfaces.hid_device import HID

from .transport import Transport

from .common import Chain

from .command_builder import DefaultInsType
from .exception import DeviceException

from .wallet import WalletPolicy
from .psbt import PSBT
from ._serialize import deser_string

try:
    from speculos.client import ApduException
except ImportError:
    # Speculos package not available, we use our own class
    class ApduException(Exception):
        def __init__(self, sw: int, data: bytes) -> None:
            super().__init__(f"Exception: invalid status 0x{sw:x}")
            self.sw = sw
            self.data = data


class TransportClient:
    def __init__(self, interface: Literal['hid', 'tcp'] = "tcp", *, server: str = "127.0.0.1", port: int = 9999, path: Optional[str] = None, hid: Optional[HID] = None, debug: bool = False):
        self.transport = Transport('hid', path=path, hid=hid, debug=debug) if interface == 'hid' else Transport(interface, server=server, port=port, debug=debug)

    def apdu_exchange(
        self, cla: int, ins: int, data: bytes = b"", p1: int = 0, p2: int = 0
    ) -> bytes:
        sw, data = self.transport.exchange(cla, ins, p1, p2, None, data)

        if sw != 0x9000:
            raise ApduException(sw, data)

        return data

    def apdu_exchange_nowait(
        self, cla: int, ins: int, data: bytes = b"", p1: int = 0, p2: int = 0
    ):
        raise NotImplementedError()

    def stop(self) -> None:
        self.transport.close()


def print_apdu(apdu_dict: dict) -> None:
    serialized_apdu = b''.join([
        apdu_dict["cla"].to_bytes(1, byteorder='big'),
        apdu_dict["ins"].to_bytes(1, byteorder='big'),
        apdu_dict["p1"].to_bytes(1, byteorder='big'),
        apdu_dict["p2"].to_bytes(1, byteorder='big'),
        len(apdu_dict["data"]).to_bytes(1, byteorder='big'),
        apdu_dict["data"]
    ])
    print(f"=> {serialized_apdu.hex()}")


def print_response(sw: int, data: bytes) -> None:
    print(f"<= {data.hex()}{sw.to_bytes(2, byteorder='big').hex()}")


@dataclass(frozen=True)
class PartialSignature:
    """Represents a partial signature returned by sign_psbt.

    It always contains a pubkey and a signature.
    The pubkey

    The tapleaf_hash is also filled if signing a for a tapscript.
    """
    pubkey: bytes
    signature: bytes
    tapleaf_hash: Optional[bytes] = None


class Client:
    def __init__(self, transport_client: TransportClient, chain: Chain = Chain.MAIN, debug: bool = False) -> None:
        self.transport_client = transport_client
        self.chain = chain
        self.debug = debug

    def _apdu_exchange(self, apdu: dict) -> Tuple[int, bytes]:
        try:
            if self.debug:
                print_apdu(apdu)

            response = self.transport_client.apdu_exchange(**apdu)
            if self.debug:
                print_response(0x9000, response)

            return 0x9000, response
        except ApduException as e:
            if self.debug:
                print_response(e.sw, e.data)

            return e.sw, e.data

    def _make_request(self, apdu: dict) -> Tuple[int, bytes]:
        return self._apdu_exchange(apdu)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.transport_client.stop()

    def stop(self) -> None:
        """Stops the transport_client."""

        self.transport_client.stop()

    def get_version(self) -> Tuple[str, str, bytes]:
        """Queries the hardware wallet for the currently running app's name, version and state flags.

        Returns
        -------
        Tuple[str, str, bytes]
            The first element is the app's name, as a short string.
            The second element is the app's version.
            The third element is a binary string representing the platform's global state (pin lock etc).
        """

        sw, response = self._make_request(
            {"cla": 0xB0, "ins": DefaultInsType.GET_VERSION, "p1": 0, "p2": 0, "data": b''})

        if sw != 0x9000:
            raise DeviceException(
                error_code=sw, ins=DefaultInsType.GET_VERSION)

        r = BytesIO(response)

        format = r.read(1)

        app_name = deser_string(r)
        app_version = deser_string(r)
        app_flags = deser_string(r)

        if format != b'\1' or app_name == b'' or app_version == b'' or app_flags == b'':
            raise DeviceException(error_code=sw, ins=DefaultInsType.GET_VERSION,
                                  message="Invalid format returned by GET_VERSION")

        return app_name.decode(), app_version.decode(), app_flags

    def get_extended_pubkey(self, path: str, display: bool = False) -> str:
        """Gets the serialized extended public key for certain BIP32 path. Optionally, validate with the user.

        Parameters
        ----------
        path : str
            BIP32 path of the public key you want.
        display : bool
            Whether you want to display address and ask confirmation on the device.

        Returns
        -------
        str
            The requested serialized extended public key.
        """

        raise NotImplementedError

    def register_wallet(self, wallet: WalletPolicy) -> Tuple[bytes, bytes]:
        """Registers a wallet policy with the user. After approval returns the wallet id and hmac to be stored on the client.

        Parameters
        ----------
        wallet : WalletPolicy
            The Wallet policy to register on the device.

        Returns
        -------
        Tuple[bytes, bytes]
            The first element the tuple is the 32-bytes wallet id.
            The second element is the hmac.
        """

        raise NotImplementedError

    def get_wallet_address(
        self,
        wallet: WalletPolicy,
        wallet_hmac: Optional[bytes],
        change: int,
        address_index: int,
        display: bool,
    ) -> str:
        """For a given wallet that was already registered on the device (or a standard wallet that does not need registration),
        returns the address for a certain `change`/`address_index` combination.

        Parameters
        ----------
        wallet : WalletPolicy
            The registered wallet policy, or a standard wallet policy.

        wallet_hmac: Optional[bytes]
            For a registered wallet, the hmac obtained at wallet registration. `None` for a standard wallet policy.

        change: int
            0 for a standard receive address, 1 for a change address. Other values are invalid.

        address_index: int
            The address index in the last step of the BIP32 derivation.

        display: bool
            Whether you want to display address and ask confirmation on the device.

        Returns
        -------
        str
            The requested address.
        """

        raise NotImplementedError

    def sign_psbt(self, psbt: Union[PSBT, bytes, str], wallet: WalletPolicy, wallet_hmac: Optional[bytes]) -> List[Tuple[int, PartialSignature]]:
        """Signs a PSBT using a registered wallet (or a standard wallet that does not need registration).

        Signature requires explicit approval from the user.

        Parameters
        ----------
        psbt : PSBT | bytes | str
            A PSBT of version 0 or 2, with all the necessary information to sign the inputs already filled in; what the
            required fields changes depending on the type of input.
            The non-witness UTXO must be present for both legacy and SegWit inputs, or the hardware wallet will reject
            signing (this will change for Taproot inputs).
            The argument can be either a `PSBT` object, or `bytes`, or a base64-encoded `str`.

        wallet : WalletPolicy
            The registered wallet policy, or a standard wallet policy.

        wallet_hmac: Optional[bytes]
            For a registered wallet, the hmac obtained at wallet registration. `None` for a standard wallet policy.

        Returns
        -------
        List[Tuple[int, PartialSignature]]
            A list of tuples returned by the hardware wallets, where each element is a tuple of:
            - an integer, the index of the input being signed;
            - an instance of `PartialSignature`.
        """

        raise NotImplementedError

    def get_master_fingerprint(self) -> bytes:
        """Gets the fingerprint of the master public key, as per BIP-32.

        Returns
        -------
        bytes
            The fingerprint of the master public key, as an array of 4 bytes.
        """

        raise NotImplementedError

    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        """
        Sign a message (bitcoin message signing).
        Signs a message using the legacy Bitcoin Core signed message format.
        The message is signed with the key at the given path.
        :param message: The message to be signed. First encoded as bytes if not already.
        :param bip32_path: The BIP 32 derivation for the key to sign the message with.
        :return: The signature
        """
        raise NotImplementedError



from .errors import MiniscriptError
from .base import DescriptorBase
from .miniscript import Miniscript
from ..hashes import tagged_hash
from ..script import Script


class TapLeaf(DescriptorBase):
    def __init__(self, miniscript=None, version=0xC0):
        self.miniscript = miniscript
        self.version = version

    def __str__(self):
        return str(self.miniscript)

    @classmethod
    def read_from(cls, s):
        ms = Miniscript.read_from(s, taproot=True)
        return cls(ms)

    def serialize(self):
        if self.miniscript is None:
            return b""
        return bytes([self.version]) + Script(self.miniscript.compile()).serialize()

    @property
    def keys(self):
        return self.miniscript.keys

    def derive(self, *args, **kwargs):
        if self.miniscript is None:
            return type(self)(None, version=self.version)
        return type(self)(
            self.miniscript.derive(*args, **kwargs),
            self.version,
        )

    def branch(self, *args, **kwargs):
        if self.miniscript is None:
            return type(self)(None, version=self.version)
        return type(self)(
            self.miniscript.branch(*args, **kwargs),
            self.version,
        )

    def to_public(self, *args, **kwargs):
        if self.miniscript is None:
            return type(self)(None, version=self.version)
        return type(self)(
            self.miniscript.to_public(*args, **kwargs),
            self.version,
        )


def _tweak_helper(tree):
    # https://github.com/bitcoin/bips/blob/master/bip-0341.mediawiki#constructing-and-spending-taproot-outputs
    if isinstance(tree, TapTree):
        tree = tree.tree
    if isinstance(tree, TapLeaf):
        # one leaf on this branch
        h = tagged_hash("TapLeaf", tree.serialize())
        return ([(tree, b"")], h)
    left, left_h = _tweak_helper(tree[0])
    right, right_h = _tweak_helper(tree[1])
    ret = [(leaf, c + right_h) for leaf, c in left] + [
        (leaf, c + left_h) for leaf, c in right
    ]
    if right_h < left_h:
        left_h, right_h = right_h, left_h
    return (ret, tagged_hash("TapBranch", left_h + right_h))


class TapTree(DescriptorBase):
    def __init__(self, tree=None):
        """tree can be None, TapLeaf or a tuple (taptree, taptree)"""
        self.tree = tree
        # make sure all keys are taproot
        for k in self.keys:
            k.taproot = True

    def __bool__(self):
        return bool(self.tree)

    def tweak(self):
        if self.tree is None:
            return b""
        _, h = _tweak_helper(self.tree)
        return h

    @property
    def keys(self):
        if self.tree is None:
            return []
        if isinstance(self.tree, TapLeaf):
            return self.tree.keys
        left, right = self.tree
        return left.keys + right.keys

    @classmethod
    def read_from(cls, s):
        c = s.read(1)
        if len(c) == 0:
            return cls()
        if c == b"{":  # more than one miniscript
            left = cls.read_from(s)
            c = s.read(1)
            if c == b"}":
                return left
            if c != b",":
                raise MiniscriptError("Invalid taptree syntax: expected ','")
            right = cls.read_from(s)
            if s.read(1) != b"}":
                raise MiniscriptError("Invalid taptree syntax: expected '}'")
            return cls((left, right))
        s.seek(-1, 1)
        ms = TapLeaf.read_from(s)
        return cls(ms)

    def derive(self, *args, **kwargs):
        if self.tree is None:
            return type(self)(None)
        if isinstance(self.tree, TapLeaf):
            return type(self)(self.tree.derive(*args, **kwargs))
        left, right = self.tree
        return type(self)((left.derive(*args, **kwargs), right.derive(*args, **kwargs)))

    def branch(self, *args, **kwargs):
        if self.tree is None:
            return type(self)(None)
        if isinstance(self.tree, TapLeaf):
            return type(self)(self.tree.branch(*args, **kwargs))
        left, right = self.tree
        return type(self)((left.branch(*args, **kwargs), right.branch(*args, **kwargs)))

    def to_public(self, *args, **kwargs):
        if self.tree is None:
            return type(self)(None)
        if isinstance(self.tree, TapLeaf):
            return type(self)(self.tree.to_public(*args, **kwargs))
        left, right = self.tree
        return type(self)(
            (left.to_public(*args, **kwargs), right.to_public(*args, **kwargs))
        )

    def __str__(self):
        if self.tree is None:
            return ""
        if isinstance(self.tree, TapLeaf):
            return str(self.tree)
        (left, right) = self.tree
        return "{%s,%s}" % (left, right)



from bitcoin_client.ledger_bitcoin import Client
from .conftest import SpeculosGlobals


def test_get_master_fingerprint(client: Client, speculos_globals: SpeculosGlobals):
    assert client.get_master_fingerprint() == speculos_globals.master_key_fingerprint



from ..base import EmbitError


class DescriptorError(EmbitError):
    pass


class MiniscriptError(DescriptorError):
    pass


class ArgumentError(MiniscriptError):
    pass


class KeyError(ArgumentError):
    pass
